import json
from typing import Any, Dict, Iterable, List, Type

from semantic_version import SimpleSpec

from bento.parser import Parser
from bento.tool import output, runner
from bento.util import fetch_line_in_file
from bento.violation import Violation

# Input example:
# {
#   "errors": [
#     {
#       "filename": "./foo.py",
#       "reason": "syntax error while parsing AST from file"
#     }
#   ],
#   "generated_at": "2019-09-17T16:56:51Z",
#   "metrics": {
#     "./bar.py": {
#       "CONFIDENCE.HIGH": 2.0,
#       "CONFIDENCE.LOW": 0.0,
#       "CONFIDENCE.MEDIUM": 0.0,
#       "CONFIDENCE.UNDEFINED": 0.0,
#       "SEVERITY.HIGH": 1.0,
#       "SEVERITY.LOW": 1.0,
#       "SEVERITY.MEDIUM": 0.0,
#       "SEVERITY.UNDEFINED": 0.0,
#       "loc": 3,
#       "nosec": 0
#     },
#     "./foo.py": {
#       "loc": 4,
#       "nosec": 0
#     },
#     "_totals": {
#       "CONFIDENCE.HIGH": 2.0,
#       "CONFIDENCE.LOW": 0.0,
#       "CONFIDENCE.MEDIUM": 0.0,
#       "CONFIDENCE.UNDEFINED": 0.0,
#       "SEVERITY.HIGH": 1.0,
#       "SEVERITY.LOW": 1.0,
#       "SEVERITY.MEDIUM": 0.0,
#       "SEVERITY.UNDEFINED": 0.0,
#       "loc": 7,
#       "nosec": 0
#     }
#   },
#   "results": [
#     {
#       "code": "1 import subprocess\n2 \n3 def do_it(cmd: str) -> None:\n4     subprocess.run(f\"bash -c {cmd}\", shell=True)\n",
#       "filename": "./bar.py",
#       "issue_confidence": "HIGH",
#       "issue_severity": "LOW",
#       "issue_text": "Consider possible security implications associated with subprocess module.",
#       "line_number": 1,
#       "line_range": [
#         1,
#         2
#       ],
#       "more_info": "https://bandit.readthedocs.io/en/latest/blacklists/blacklist_imports.html#b404-import-subprocess",
#       "test_id": "B404",
#       "test_name": "blacklist"
#     },
#     {
#       "code": "3 def do_it(cmd: str) -> None:\n4     subprocess.run(f\"bash -c {cmd}\", shell=True)\n",
#       "filename": "./bar.py",
#       "issue_confidence": "HIGH",
#       "issue_severity": "HIGH",
#       "issue_text": "subprocess call with shell=True identified, security issue.",
#       "line_number": 4,
#       "line_range": [
#         4
#       ],
#       "more_info": "https://bandit.readthedocs.io/en/latest/plugins/b602_subprocess_popen_with_shell_equals_true.html",
#       "test_id": "B602",
#       "test_name": "subprocess_popen_with_shell_equals_true"
#     }
#   ]
# }

BANDIT_TO_BENTO = {
    "B101": "assert-used",
    "B102": "exec-used",
    "B103": "set-bad-file-permissions",
    "B104": "hardcoded-bind-all-interfaces",
    "B105": "hardcoded-password-string",
    "B106": "hardcoded-password-funcarg",
    "B107": "hardcoded-password-default",
    "B108": "hardcoded-tmp-directory",
    "B110": "try-except-pass",
    "B112": "try-except-continue",
    "B201": "flask-debug-true",
    "B301": "pickle-used",
    "B303": "md5",
    "B307": "eval-used",
    "B308": "django-mark-safe-used",
    "B309": "https-connection",
    "B313": "xml-bad-cElementTree",
    "B314": "xml-bad-ElementTree",
    "B315": "xml-bad-expatreader",
    "B316": "xml-bad-expatbuilder",
    "B317": "xml-bad-sax",
    "B318": "xml-bad-minidom",
    "B319": "xml-bad-pulldom",
    "B320": "xml-bad-etree",
    "B321": "ftplib",
    "B310": "urllib-urlopen",
    "B311": "random",
    "B322": "python2-input",
    "B323": "unverified-context",
    "B324": "hashlib-new-insecure-functions",
    "B325": "tempnam",
    "B401": "import-telnetlib",
    "B402": "import-ftplib",
    "B403": "import-pickle",
    "B404": "import-subprocess",
    "B405": "import-xml-etree",
    "B406": "import-xml-sax",
    "B407": "import-xml-expat",
    "B408": "import-xml-minidom",
    "B409": "import-xml-pulldom",
    "B410": "import-lxml",
    "B411": "import-xmlrpclib",
    "B412": "import-httpoxy",
    "B413": "import-pycrypto",
    "B501": "request-with-no-cert-validation",
    "B502": "ssl-with-bad-version",
    "B503": "ssl-with-bad-defaults",
    "B504": "ssl-with-no-version",
    "B505": "weak-cryptographic-key",
    "B506": "yaml-load",
    "B507": "ssh-no-host-key-verification",
    "B601": "paramiko-calls",
    "B602": "subprocess-popen-with-shell-equals-true",
    "B603": "subprocess-without-shell-equals-true",
    "B604": "any-other-function-with-shell-equals-true",
    "B605": "start-process-with-a-shell",
    "B606": "start-process-with-no-shell",
    "B607": "start-process-with-partial-path",
    "B608": "hardcoded-sql-expressions",
    "B609": "linux-commands-wildcard-injection",
    "B610": "django-extra-used",
    "B611": "django-rawsql-used",
    "B701": "jinja2-autoescape-false",
    "B702": "use-of-mako-templates",
    "B703": "django-mark-safe",
}


class BanditParser(Parser[str]):
    SEVERITY = {"LOW": 0, "MEDIUM": 1, "HIGH": 2}
    LINE_NO_CHARS = "0123456789"

    def __error_to_violation(self, error: Dict[str, Any]) -> Violation:
        return Violation(
            check_id="error",
            tool_id=BanditTool.TOOL_ID,
            path=self.trim_base(error["filename"]),
            severity=2,
            line=0,
            column=0,
            message=error["reason"],
            syntactic_context="",
            link=None,
        )

    def __result_to_violation(self, result: Dict[str, Any]) -> Violation:
        path = self.trim_base(result["filename"])
        link = result.get("more_info", None)

        # Remove bandit line numbers, empty lines, and leading / trailing whitespace
        bandit_source = result["code"].rstrip()  # Remove trailing whitespace

        test_id = result["test_id"]
        check_id = BANDIT_TO_BENTO.get(test_id, test_id)

        line_range = result["line_range"]

        def in_line_range(bandit_code_line: str) -> bool:
            # Check if string with format `3 def do_it(cmd: str) -> None:`
            # starts with line number that is within reported line_range
            # of finding
            for idx, ch in enumerate(bandit_code_line):
                if not ch.isdigit():
                    num = int(bandit_code_line[:idx])
                    return num in line_range
            return False

        # bandit might include extra lines before and after
        # a finding. Filter those out and filter out line numbers
        lines = [
            s.lstrip(BanditParser.LINE_NO_CHARS).rstrip()
            for s in bandit_source.split("\n")
            if in_line_range(s)
        ]
        nonempty = [l for l in lines if l]
        source = "\n".join(nonempty)

        if source == "" and result["line_number"] != 0:
            source = (
                fetch_line_in_file(self.base_path / path, result["line_number"])
                or "<no source found>"
            )

        return Violation(
            check_id=check_id,
            tool_id=BanditTool.TOOL_ID,
            path=path,
            line=result["line_number"],
            column=0,
            message=result["issue_text"],
            severity=BanditParser.SEVERITY.get(result["issue_severity"], 1),
            syntactic_context=source,
            link=link,
        )

    def parse(self, tool_output: str) -> List[Violation]:
        results: Dict[str, List[Dict[str, Any]]] = json.loads(tool_output)
        errors = [self.__error_to_violation(e) for e in results.get("errors", [])]
        violations = [self.__result_to_violation(r) for r in results.get("results", [])]
        return errors + violations


class BanditTool(runner.Python, output.Str):
    TOOL_ID = "bandit"  # to-do: versioning?
    VENV_DIR = "bandit"
    PROJECT_NAME = "Python"
    PACKAGES = {"bandit": SimpleSpec("~=1.6.2")}

    @property
    def parser_type(self) -> Type[Parser]:
        return BanditParser

    @classmethod
    def tool_id(self) -> str:
        return BanditTool.TOOL_ID

    @classmethod
    def tool_desc(cls) -> str:
        return "Finds common security issues in Python code"

    @classmethod
    def venv_subdir_name(cls) -> str:
        return BanditTool.VENV_DIR

    @property
    def project_name(self) -> str:
        return BanditTool.PROJECT_NAME

    def run(self, paths: Iterable[str]) -> str:
        cmd = [
            "python",
            str(self.venv_dir() / "bin" / "bandit"),
            "-f",
            "json",
            "-r",
            *paths,
        ]
        return self.venv_exec(cmd, check_output=False)



import itertools
import shutil
import textwrap
from typing import Collection, List

import click

from bento.formatter.base import FindingsMap, Formatter
from bento.util import ANSI_WIDTH, PRINT_WIDTH, render_link
from bento.violation import Violation


class Stylish(Formatter):
    """
    Mimics the eslint "stylish" formatter
    """

    TERM_WIDTH, _ = shutil.get_terminal_size((PRINT_WIDTH, 0))

    @staticmethod
    def __print_path(path: str) -> str:
        return path

    def __print_violation(self, violation: Violation) -> List[str]:
        message_len = max(Stylish.TERM_WIDTH - 50, 40)

        line = f"{violation.line:>4d}"
        col = f"{violation.column:<2d}"
        tool_id = f"{violation.tool_id:<14s}"
        rule = f"{violation.check_id:s}"

        message = textwrap.wrap(violation.message.strip(), message_len)

        if not violation.link:
            link = rule
        else:
            link = render_link(rule, violation.link)

        out = [
            f"{line}:{col} {click.style(message[0], dim=True):<{message_len + 2 * ANSI_WIDTH}s} {tool_id} {link}"
        ]
        for m in message[1:]:
            out.append(
                f"        {click.style(m, dim=True):<{message_len + 2 * ANSI_WIDTH}s}"
            )

        return out

    def dump(self, findings: FindingsMap) -> Collection[str]:
        violations = self.by_path(findings)
        lines = []

        ordered: List[Violation] = sorted(violations, key=Formatter.path_of)
        for path, vv in itertools.groupby(ordered, Formatter.path_of):
            lines.append(Stylish.__print_path(path))
            for v in sorted(vv, key=lambda v: (v.line, v.column, v.message)):
                lines += self.__print_violation(v)
            lines.append("")

        return lines



import itertools
from typing import Collection, List, Optional

import attr
import click

from bento.formatter.base import FindingsMap, Formatter
from bento.util import PRINT_WIDTH, render_link
from bento.violation import Violation


@attr.s(auto_attribs=True)
class Hit:
    check_id: str
    link: Optional[str]
    count: int  # Number of findings for this check


@attr.s(auto_attribs=True)
class ToolHits:
    tool_id: str
    hits: List[Hit]  # Ordered list of Hit objects for this tool
    max_count = 0
    max_check_width = 0

    def __attrs_post_init__(self) -> None:
        if self.hits:
            self.max_count = max(h.count for h in self.hits)
            self.max_check_width = max(len(h.check_id) for h in self.hits)


class Histo(Formatter):
    """Formats output as a histogram."""

    MAX_HISTO_PER_TOOL = 5
    MIN_CHECK_WIDTH = 10
    MAX_CHECK_WIDTH = 35
    OTHER = "Other"

    @staticmethod
    def _render_bar(count: int, max_count: int, bar_width: int) -> str:
        return "▕" + "".ljust(int(round(bar_width * count / max_count)), "▬")

    def _max_bars_per_tool(self) -> int:
        return int(self.config.get("bars-per-tool", Histo.MAX_HISTO_PER_TOOL))

    def _render_hit(self, hit: Hit, max_count: int, check_width: int) -> str:
        bar_width = (
            PRINT_WIDTH - check_width - 10
        )  # 5 for check, 5 for fixed characters
        check_str = click.style(
            render_link(
                hit.check_id[:check_width],
                href=hit.link,
                print_alternative=False,
                width=check_width,
            )
        )
        count_str = click.style(f"{hit.count:>5d}", dim=True)
        bar = click.style(Histo._render_bar(hit.count, max_count, bar_width), dim=True)
        return f"  {check_str} {count_str}{bar}"

    def _tool_hits(self, tool_id: str, violations: Collection[Violation]) -> ToolHits:
        """
        Collects top hits for a single tool.
        """
        keyfunc = lambda v: v.check_id  # noqa
        sorted_violations = sorted(violations, key=keyfunc)
        counts = sorted(
            (
                Hit(check_id.strip(), next(vv).link, 1 + sum(1 for _ in vv))
                for check_id, vv in itertools.groupby(sorted_violations, keyfunc)
            ),
            key=(lambda hit: hit.count),
            reverse=True,
        )
        top = counts[0 : self._max_bars_per_tool()]
        n_top = sum(h.count for h in top)

        # Add remaining hits as "Other" if more hits than max_bars_per_tool
        if len(counts) > len(top):
            top.append(Hit(Histo.OTHER, None, len(violations) - n_top))

        return ToolHits(tool_id, top)

    def _all_hits(self, findings: FindingsMap) -> List[ToolHits]:
        """
        For each tool, returns an ordered list of (ToolId, CheckData, max(Count)).

        Each list is the top checks by finding count, plus an "Other" row if more than MAX_HISTO_PER_TOOL checks fired.
        """
        return [
            self._tool_hits(tool_id, violations)
            for tool_id, violations in findings.items()
        ]

    def dump(self, findings: FindingsMap) -> Collection[str]:
        """
        Prints an ordered histogram of "top hits" per tool.

        E.g.:
            r2c.eslint:
              no-undef                  16015 ++++++++++++++++++++++++++++++++++++++++++++++++++
              no-var                     7880 ++++++++++++++++++++++++
              @typescript-eslint/explic  4864 +++++++++++++++
              object-shorthand           1399 ++++
              @typescript-eslint/no-var   157
              Other                      1178 +++

            r2c.flake8:
              B008:                       453 +
              A003                        103
              F405                         62
              B007                         56
              B006                         52
              Other                       116
        """
        if not findings:
            return []

        all_hits = self._all_hits(findings)
        max_count = max(h.max_count for h in all_hits)
        check_width = max(h.max_check_width for h in all_hits)
        check_width = max(
            min(check_width, Histo.MAX_CHECK_WIDTH), Histo.MIN_CHECK_WIDTH
        )
        out = []
        for tool_hits in sorted(all_hits, key=(lambda h: h.max_count), reverse=True):
            if tool_hits.hits:
                out.append(click.style(f"{tool_hits.tool_id}:", bold=True))
                for hit in tool_hits.hits:
                    out.append(self._render_hit(hit, max_count, check_width))
                out.append("")
        return out



import logging
from pathlib import Path
from threading import Lock
from typing import Any, Dict, Optional, Union

import attr
import yaml

import bento.constants as constants
import bento.git
from bento.error import NoConfigurationException, NoIgnoreFileException
from bento.run_cache import RunCache

DEFAULT_CONFIG_PATH = Path(__file__).parent / "configs" / "headless.yml"

DEFAULT_IGNORE_PATH = Path(__file__).parent / "configs" / constants.IGNORE_FILE_NAME


def _clean_path(path: Optional[Union[str, Path]]) -> Optional[Path]:
    # The resolve here is important, since flake8 seems to have weird behavior
    # regarding finding unused imports if the path is not fully-resolved.
    return Path(path).resolve() if path else None


@attr.s
class BaseContext:
    # The path to the directory that contains the .bento dir and .bento.yml
    # file. Tools will also be run from here.
    base_path: Path = attr.ib(converter=_clean_path)  # type: ignore
    cache_path: Path = attr.ib(converter=_clean_path, default=None)  # type: ignore
    is_init = attr.ib(type=bool, default=False)
    email = attr.ib(type=Optional[str], default=None)
    headless = attr.ib(type=bool, default=False)
    prog_name = attr.ib(type=str, default=None)
    _config = attr.ib(type=Dict[str, Any], default=None)
    _resource_path = attr.ib(type=Path, default=None)
    _cache = attr.ib(type=RunCache, default=None, init=False)
    _ignore_lock = attr.ib(type=Lock, factory=Lock, init=False)

    @base_path.default
    def _find_base_path(self) -> Path:
        """Find the path to the nearest containing directory with bento config.

        This starts at the current directory, then recurses upwards looking for
        a directory with the necessary config file. This function will not recurse
        at or beyond the git project root or the user's home directory.

        If one isn't found, returns the current working directory. This
        behavior is so that you can construct a Context in a directory that
        doesn't have Bento set up, then use the config_path and such to figure
        out where to do the initialization.

        """
        cwd = Path.cwd()

        repo_root = None
        repo_root_obj = bento.git.repo()
        if repo_root_obj is not None:
            repo_root = Path(repo_root_obj.working_tree_dir)

        for base_path in [cwd, *cwd.parents]:
            config_path = base_path / constants.RESOURCE_PATH
            if config_path == constants.GLOBAL_RESOURCE_PATH:  # Don't go past user home
                break
            if (config_path / constants.CONFIG_FILE_NAME).is_file():
                return base_path
            # Stop at root of git repository
            if base_path == repo_root:
                return base_path
        return cwd

    @property
    def config_path(self) -> Path:
        return self.base_path / constants.RESOURCE_PATH / constants.CONFIG_FILE_NAME

    @property
    def resource_path(self) -> Path:
        if not self._resource_path:
            self._resource_path = self.base_path / constants.RESOURCE_PATH
        return self._resource_path

    @property
    def baseline_file_path(self) -> Path:
        return self.resource_path / constants.ARCHIVE_FILE_NAME

    @property
    def ignore_file_path(self) -> Path:
        return self.base_path / constants.IGNORE_FILE_NAME

    @property
    def gh_actions_file_path(self) -> Path:
        return self.base_path / constants.GH_ACTIONS_FILE_NAME

    def pretty_path(self, path: Path) -> Path:
        try:
            return path.relative_to(self.base_path)
        except ValueError:
            return path

    def __attrs_post_init__(self) -> None:
        # We need to make sure the resource directory exists prior to creating the FileIgnore object,
        # otherwise it won't be detected in its directory scan.
        self.resource_path.mkdir(parents=True, exist_ok=True)

    @property
    def config(self) -> Dict[str, Any]:
        if self._config is None:
            self._config = self._open_config()
        return self._config

    @config.setter
    def config(self, config: Dict[str, Any]) -> None:
        self._write_config(config)
        self._config = config

    @property
    def autorun_is_blocking(self) -> bool:
        """
        Returns whether `bento check --staged-only` should block commits
        """
        return self.config.get("autorun", {}).get("block", False)

    @property
    def cache(self) -> RunCache:
        if self._cache is None:
            cp = self.cache_path or (self.resource_path / constants.CACHE_PATH)
            self._cache = RunCache(cache_dir=cp)
        return self._cache

    def validate_init(self) -> None:
        if self.headless:
            return

        # Fail out if not configured
        if not self.config_path.exists():
            raise NoConfigurationException()

        # Fail out if no .bentoignore
        if not self.ignore_file_path.exists():
            raise NoIgnoreFileException(self.ignore_file_path)

    def _open_config(self) -> Dict[str, Any]:
        """
        Opens this project's configuration file
        """
        logging.info(f"Loading bento configuration from {self.config_path}")
        config_path = (
            self.config_path if self.config_path.exists() else DEFAULT_CONFIG_PATH
        )
        with config_path.open() as yaml_file:
            return yaml.safe_load(yaml_file)

    def _write_config(self, config: Dict[str, Any]) -> None:
        """
        Overwrites this project's configuration file
        """
        if self.config_path.exists():
            logging.info(f"Writing bento configuration to {self.config_path}")
            with self.config_path.open("w") as yaml_file:
                yaml.safe_dump(config, yaml_file)



from bento.renderer import (
    Box,
    Confirm,
    Content,
    Echo,
    Link,
    Multi,
    Newline,
    Processors,
    Progress,
    Steps,
    Sub,
    Text,
    Warn,
)

from . import ci as ci_content


def _step_item(desc: str, cmd: str, nl: bool = True) -> Content:
    """
    Echoes

      desc․․․․․․․․․․․․․․․․․․․․․․․․․․․․․․․․․․ $ cmd

    with styling.

    If 'nl' is False, does not emit a newline at the end of the line
    """
    return Multi(
        [
            "  ",
            Text(desc, processor=Processors.ljust(-40, "․"), style={"dim": True}),
            f" $ {cmd}\n",
        ]
    )


class InstallConfig:
    install = Progress(
        content=Multi(
            [
                "Creating default configuration at ",
                Sub(0, style={"bold": True, "dim": True}),
            ]
        ),
        extra=12,
    )


class InstallIgnore:
    install = Progress(
        content=Multi(
            [
                "Creating default ignore file at ",
                Sub(0, style={"bold": True, "dim": True}),
            ]
        ),
        extra=12,
    )


class InstallAutorun:
    install = Progress(
        Multi(
            [
                "Enabling autorun (see $ ",
                Text("bento enable autorun --help", style={"dim": True, "bold": True}),
                Text(")", style={"dim": True}),
            ]
        ),
        extra=20,
    )


class InstallCI:
    pitch = Steps(
        Echo(
            Multi(
                [
                    "Bento can configure a GitHub Action to automatically check your team's pull requests. ",
                    ci_content.Install.learn_more,
                ],
                processor=Processors.wrap(),
            )
        ),
        Newline(),
    )

    confirm = Confirm("Do you want to configure Bento's GitHub Action now?")
    after_confirm = Newline()

    progress = ci_content.Install.progress
    finalize_ci = ci_content.Install.finalize_ci


class InstallTools:
    install = Echo("Installing tools:\n")


class Clean:
    tools = Steps(Echo("Reinstalling tools due to passed --clean flag."), Newline())
    check = Warn(
        """
Removing archive due to passed --clean flag.
"""
    )


class Identify:
    success = Steps(
        Newline(),
        Echo(Multi(["Bento initialized for ", Sub(0, style={"bold": True})])),
        Newline(),
    )
    failure = Steps(
        Newline(),
        Warn(
            "Bento can't automatically identify this project. Please manually configure this project:"
        ),
        Echo(
            Multi(
                [
                    _step_item("configure a tool manually", "bento enable tool TOOL"),
                    _step_item("list available tools", "bento enable tool --help"),
                ]
            )
        ),
    )


class Finish:
    body = Steps(
        Box("Thank You"),
        Echo(
            Text(
                "From all of us at r2c, thank you for trying Bento! We can’t wait to hear what you think.",
                processor=Processors.wrap(),
            )
        ),
        Newline(),
        Echo(
            Multi(
                [
                    "Help and feedback: ",
                    Text("Reach out to us at ", style={"dim": True}),
                    "support@r2c.dev",
                    Text(" or file an issue on ", style={"dim": True}),
                    "GitHub",
                    Text(". We’d love to hear from you!", style={"dim": True}),
                ],
                processor=Processors.wrap_link(
                    [
                        Link("support@r2c.dev", "mailto:support@r2c.dev"),
                        Link("GitHub", "https://github.com/returntocorp/bento/issues"),
                    ],
                    extra=16,
                ),
            )
        ),
        Newline(),
        Echo(
            Multi(
                [
                    "Community: ",
                    Text("Join ", style={"dim": True}),
                    "#bento",
                    Text(
                        " on our community Slack. Get support, talk with other users, and share feedback.",
                        style={"dim": True},
                    ),
                ],
                processor=Processors.wrap_link(
                    [
                        Link(
                            "#bento",
                            "https://join.slack.com/t/r2c-community/shared_invite/enQtNjU0NDYzMjAwODY4LWE3NTg1MGNhYTAwMzk5ZGRhMjQ2MzVhNGJiZjI1ZWQ0NjQ2YWI4ZGY3OGViMGJjNzA4ODQ3MjEzOWExNjZlNTA",
                        )
                    ],
                    extra=16,
                ),
            )
        ),
        Newline(),
        Echo("Go forth and write great code! To use Bento:"),
        Echo(
            Multi(
                [
                    _step_item("commit code", "git commit"),
                    _step_item(
                        "get help for a command", "bento [COMMAND] --help", nl=False
                    ),
                ]
            )
        ),
    )


class Start:
    banner = Steps(
        Box("Bento Initialization"),
        Echo(
            Text(
                "Bento configures itself for personal use by default. This means that it:",
                processor=Processors.wrap(),
            )
        ),
        Newline(),
        Echo(
            Text(
                "1. Automatically checks for issues introduced by your code, as you commit it",
                processor=Processors.wrap(),
            )
        ),
        Echo(
            Text(
                "2. Only affects you; it won’t change anything for other project contributors",
                processor=Processors.wrap(),
            )
        ),
        Newline(),
        Echo(
            Text(
                "Learn more about personal and team use at bento.dev/workflows.",
                processor=Processors.wrap_link(
                    [Link("bento.dev/workflows", "https://bento.dev/workflows")]
                ),
            )
        ),
        Newline(),
    )
    confirm = Steps(Confirm("Press ENTER to add Bento to this project"), Newline())



import itertools
import shutil
import sys
import textwrap
from typing import Collection, List

import click

from bento.formatter.base import FindingsMap, Formatter
from bento.util import PRINT_WIDTH, render_link
from bento.violation import Violation


class Clippy(Formatter):
    """
    Mimics the clippy formatter
    """

    PIPE = click.style("│", dim=True)
    CONTEXT_HEADER = f"     {click.style('╷', dim=True)}"
    CONTEXT_FOOTER = f"     {click.style('╵', dim=True)}"
    NOTE_LEADER = f"     = "
    LEADER_LEN = len(NOTE_LEADER)
    MESSAGE_OVERFLOW_LEADER = "".ljust(len(NOTE_LEADER), " ")
    MIN_MESSAGE_LEN = 40

    @staticmethod
    def _print_path(path: str, line: int, col: int) -> str:
        fpos = click.style(f"{path}:{line}")
        return f"     > {fpos}"

    def _print_violation(self, violation: Violation, max_message_len: int) -> List[str]:
        message_lines = textwrap.wrap(violation.message.strip(), width=max_message_len)

        out = []

        # Strip so trailing newlines are not printed out
        stripped = violation.syntactic_context.rstrip()
        context = [line.rstrip() for line in stripped.split("\n")]

        if stripped:
            out.append(Clippy.CONTEXT_HEADER)
            for offset, line in enumerate(context):
                line_no = click.style(f"{violation.line + offset:>4d}", dim=True)
                out.append(f" {line_no}{Clippy.PIPE}   {line}")
            out.append(Clippy.CONTEXT_FOOTER)

        out.append(f"{Clippy.NOTE_LEADER}{click.style(message_lines[0], dim=True)}")
        for mline in message_lines[1:]:
            out.append(
                f"{Clippy.MESSAGE_OVERFLOW_LEADER}{click.style(mline, dim=True)}"
            )

        return out

    def _print_error_message(self, violation: Violation) -> str:
        tool_id = f"{violation.tool_id}".strip()
        rule = f"{violation.check_id}".strip()

        link = render_link(rule, violation.link)

        return f"{Clippy.BOLD}{tool_id} {link}{Clippy.END}"

    def dump(self, findings: FindingsMap) -> Collection[str]:
        if not findings:
            return []

        lines = []
        violations = self.by_path(findings)
        max_message_len = min(
            max((len(v.message) for v in violations), default=0),
            PRINT_WIDTH - Clippy.LEADER_LEN,
        )

        if sys.stdout.isatty():
            terminal_width, _ = shutil.get_terminal_size((PRINT_WIDTH, 0))
            max_message_len = max(
                min(max_message_len, terminal_width - Clippy.LEADER_LEN),
                Clippy.MIN_MESSAGE_LEN,
            )

        ordered: List[Violation] = sorted(violations, key=Formatter.path_of)

        for path, vv in itertools.groupby(ordered, Formatter.path_of):
            for v in sorted(vv, key=lambda v: (v.line, v.column, v.message)):
                lines.append(self._print_error_message(v))
                lines.append(Clippy._print_path(path, v.line, v.column))

                lines += self._print_violation(v, max_message_len)
                lines.append("")

        return lines



import itertools
import os
from hashlib import sha256
from typing import Any, Dict, Iterable, List, Optional, Tuple

import bento.git
from bento.constants import ARGS_TO_EXCLUDE_FROM_METRICS
from bento.util import read_global_config
from bento.violation import Violation

USAGE_SALT = "566F73BE939D2383A93F9D5759328".encode()
RESULTS_SALT = "266DD8584C42A12FC9447B1F9AFC6".encode()


def __hash_sha256(data: Optional[str], salt: bytes) -> Optional[str]:
    """ Get SHA256 of data
    """
    if data is None:
        return None

    hsh = sha256(salt)
    hsh.update(data.encode())
    return hsh.hexdigest()


def __get_filtered_violation_count(violations: Iterable[Violation]) -> int:
    return sum(1 for v in violations if v.filtered)


def __get_aggregate_violations(violations: List[Violation]) -> List[Dict[str, Any]]:
    """Returns count of violation per file, per check_id"""

    def grouping(v: Violation) -> Tuple[str, str]:
        return v.path, v.check_id

    out = []
    for k, v in itertools.groupby(sorted(violations, key=grouping), grouping):
        p, rid = k
        out.append(
            {
                "path_hash": __hash_sha256(p, RESULTS_SALT),
                "check_id": rid,
                "count": sum(1 for _ in v),
                "filtered_count": __get_filtered_violation_count(v),
            }
        )
    return out


def _infer_ci_provider() -> str:
    """Returns used CI provider based on environment."""
    if os.getenv("BENTO_ACTION") == "true":
        return "github-actions/bento-action"

    # https://help.github.com/en/actions/configuring-and-managing-workflows/using-environment-variables
    elif os.getenv("GITHUB_ACTIONS") == "true":
        return "github-actions"

    # https://circleci.com/docs/2.0/env-vars/#built-in-environment-variables
    elif os.getenv("CIRCLECI") == "true":
        return "circleci"

    # https://docs.travis-ci.com/user/environment-variables/#default-environment-variables
    elif os.getenv("TRAVIS") == "true":
        return "travis"

    # https://docs.gitlab.com/ee/ci/variables/predefined_variables.html
    elif os.getenv("GITLAB_CI") == "true":
        return "gitlab-ci"

    elif os.getenv("CI"):
        return "unknown"

    else:
        return "none"


def violations_to_metrics(
    tool_id: str, timestamp: str, violations: List[Violation], ignores: List[str]
) -> List[Dict[str, Any]]:
    # NOTE: Do not calculate url() and commit() on a per-item basis.
    # Doing so causes bento check to take many unnecessary seconds.
    url = bento.git.url()
    commit = bento.git.commit()
    return [
        {
            "tool": tool_id,
            "timestamp": timestamp,
            "hash_of_repository": __hash_sha256(url, RESULTS_SALT),
            "hash_of_commit": __hash_sha256(commit, RESULTS_SALT),
            "ignored_rules": ignores,
            **aggregates,
        }
        for aggregates in __get_aggregate_violations(violations)
    ]


def read_user_email() -> Optional[str]:
    # email guaranteed to exist because we check it for every command
    global_config = read_global_config()
    if not global_config:
        return None

    return global_config.get("email")


def command_metric(
    command: str,
    email: str,
    timestamp: str,
    command_kwargs: Dict[str, Any],
    exit_code: int,
    duration: float,
    exception_name: Optional[str],
    user_duration: Optional[float] = None,
) -> List[Dict[str, Any]]:

    # remove kwargs that contain sensitive data from metrics
    kwargs_to_exclude = ARGS_TO_EXCLUDE_FROM_METRICS.get(command, set())
    command_kwargs = dict(
        (k, v) for k, v in command_kwargs.items() if k not in kwargs_to_exclude
    )

    d = {
        "timestamp": timestamp,
        "duration": duration,
        "user_duration": user_duration,
        "exit_code": exit_code,
        "hash_of_repository": __hash_sha256(bento.git.url(), USAGE_SALT),
        "email": email,
        "hash_of_commit": __hash_sha256(bento.git.commit(), USAGE_SALT),
        "command": command,
        "command_kwargs": command_kwargs,
        "is_ci": bool(os.environ.get("CI", False)),
        "ci_provider": _infer_ci_provider(),
        "exception_name": exception_name,
    }
    return [d]



import logging
import os
import shutil
import sys
import tempfile
from pathlib import Path
from typing import Tuple

import click

from bento.decorators import with_metrics
from bento.util import append_text_to_file, echo_error, echo_success, file_has_text

ZSH_TEXT = 'eval "$(_BENTO_COMPLETE=source_zsh bento)"'
BASH_TEXT = 'eval "$(_BENTO_COMPLETE=source bento)"'

SUPPORTED = {"bash": (".bashrc", BASH_TEXT), "zsh": (".zshrc", ZSH_TEXT)}
VALID = ",".join(SUPPORTED.keys())


def _validate_shell() -> Tuple[Path, str]:
    """
    Gets the profile file and completion text for the current shell

    :raises SystemExit: If not in a valid shell
    """
    shell_env = os.environ.get("SHELL")
    logging.info(f"Using shell {shell_env}")
    if not shell_env:
        echo_error(
            f"This command must be executed within one of the {VALID} shells. (Currently not in a shell)."
        )
        sys.exit(1)

    shell = shell_env.split("/")[-1]
    if shell not in SUPPORTED:
        echo_error(
            f"This command must be executed within one of the {VALID} shells. (Currently using {shell})."
        )
        sys.exit(1)

    profile, text = SUPPORTED[shell]
    path = Path.home() / profile

    return path, text


@click.command(name="autocomplete")
@with_metrics
def install_autocomplete(quiet: bool = False) -> None:
    """
    Enable tab autocompletion in your shell profile.
    """
    path, text = _validate_shell()

    if not path.exists():
        path.touch(mode=0o644)

    if not file_has_text(path, text):
        append_text_to_file(path, text)

    if not quiet:
        echo_success(
            f"Tab autocompletion for Bento has been added to {path}. Please run `source {path}`."
        )


@click.command(name="autocomplete")
@with_metrics
def uninstall_autocomplete() -> None:
    """
    Remove tab autocompletion from your shell profile.
    """
    path, text = _validate_shell()

    # Removes "{text}\n" from file; however, it leaves the preceding newline.
    # This is simply because the logic got complicated otherwise.

    omit_next_newline = False
    swap_fd, swap_path = tempfile.mkstemp(prefix="path_", text=True)
    try:
        with path.open() as profile_fd:
            for line in profile_fd:
                if line.strip() == text:
                    omit_next_newline = True
                elif omit_next_newline and line == "\n":
                    omit_next_newline = False
                else:
                    omit_next_newline = False
                    os.write(swap_fd, line.encode())
        os.fsync(swap_fd)
        shutil.copyfile(swap_path, path)
    finally:
        os.close(swap_fd)
        os.remove(swap_path)

    echo_success(f"Tab autocompletion for Bento has been removed from {path}.")



import logging
import multiprocessing.synchronize
import sys
import threading
import time
import traceback
from contextlib import contextmanager
from enum import Enum
from functools import partial
from multiprocessing import Lock
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import Collection, Iterable, Iterator, List, Optional, Tuple, Union

import attr
import click
from tqdm import tqdm

import bento.result
import bento.util
from bento.error import NoToolsConfiguredException
from bento.result import Baseline
from bento.tool import Tool
from bento.violation import Violation

DONE_BAR_VALUE = 30
BAR_UPDATE_INTERVAL = 0.1

SLOW_RUN_SECONDS = 60  # Number of seconds before which a "slow run" warning is printed

ToolResults = Union[List[Violation], Exception]
RunResults = Tuple[str, ToolResults]

START_RUN_BAR_VALUE = int(DONE_BAR_VALUE / 5)


class RunStep(Enum):
    BASELINE = 1
    CHECK = 2


@attr.s
class Runner:
    paths = attr.ib(type=List[Path])
    use_cache = attr.ib(type=bool)
    skip_setup = attr.ib(type=bool, default=False)
    show_bars = attr.ib(type=bool, default=True)
    install_only = attr.ib(type=bool, default=False)
    _lock = attr.ib(type=multiprocessing.synchronize.Lock, factory=Lock, init=False)
    _bars = attr.ib(type=List[tqdm], factory=list, init=False)
    _run = attr.ib(type=List[bool], factory=list, init=False)
    _done = attr.ib(type=bool, default=False, init=False)

    def __attrs_post_init__(self) -> None:
        self.show_bars = self.show_bars and sys.stderr.isatty()

    def _update_bars(self, ix: int, lower: int, upper: int) -> None:
        """
        Increments a progress bar
        """
        bar_value = lower
        keep_going = True
        bar = self._bars[ix]
        while keep_going:
            with self._lock:
                keep_going = self._run[ix]
                if bar_value < upper - 1:
                    bar_value += 1
                    bar.update(1)
                else:
                    bar_value = lower
                    bar.update(lower - upper + 1)
            time.sleep(BAR_UPDATE_INTERVAL)
        with self._lock:
            bar.update(upper - bar_value)

    def _setup_bars(self, indices_and_tools: List[Tuple[int, Tool]]) -> None:
        """
        Constructs progress bars for all tools
        """
        self._bars = [
            tqdm(
                total=DONE_BAR_VALUE,
                position=ix,
                ascii="□■",
                mininterval=BAR_UPDATE_INTERVAL,
                desc=tool.tool_id().ljust(
                    bento.util.PRINT_WIDTH - 34, bento.util.LEADER_CHAR
                ),
                ncols=bento.util.PRINT_WIDTH - 2,
                bar_format=click.style("{desc:s}|{bar}| {elapsed}{postfix}", dim=True),
                leave=False,
            )
            for ix, tool in indices_and_tools
        ]

    @contextmanager
    def _updating_bar(
        self,
        bar: Optional[tqdm],
        ix: int,
        min_value: int,
        max_value: int,
        start_text: str,
        done_text: str,
    ) -> Iterator[None]:
        """
        Runs a block with an updating progress bar

        :param bar: The progress bar (or None if no progress bar is displayed for this item)
        :param ix: The progress bar index
        :param min_value: Min value to display
        :param max_value: Max value to display
        """
        th: Optional[threading.Thread] = None
        self._run[ix] = True
        if bar:
            with self._lock:
                bar.set_postfix_str(start_text)
            th = threading.Thread(
                name=f"update_{ix}",
                target=partial(self._update_bars, ix, min_value, max_value),
            )
            th.start()
        try:
            yield
        finally:
            self._run[ix] = False
            if th:
                th.join()
            if bar:
                with self._lock:
                    bar.n = max_value
                    bar.update(0)
                    bar.set_postfix_str(done_text)

    def _echo_slow_run(self) -> None:
        """
        Echoes a warning to the screen that Bento is taking longer than expected.

        Wipes and resets any progress bars.
        """
        before = time.monotonic()

        while not self._done and time.monotonic() - before < SLOW_RUN_SECONDS:
            time.sleep(BAR_UPDATE_INTERVAL)

        if not self._done:
            with self._lock:
                if self.show_bars:
                    wipe_line = "".ljust(bento.util.PRINT_WIDTH, " ")

                    click.echo(
                        "\x1b[0G", nl=False, err=True
                    )  # Reset cursor to beginning of line
                    click.echo("\x1b[s", nl=False, err=True)  # Save cursor state
                    for _ in range(len(self._bars) - 1):
                        click.echo(wipe_line, err=True)
                    # Handling last line separately prevents screen roll
                    click.echo(wipe_line, nl=False, err=True)
                    click.echo("\x1b[u", nl=False, err=True)  # Retrieve cursor state

                click.secho(
                    bento.util.wrap(
                        f"Bento is taking longer than expected, which may mean it’s checking build or dependency "
                        f"code. This is often unintentional or unexpected.",
                        extra=-4,
                    )
                    + "\n\n"
                    + bento.util.wrap(
                        "Please make sure all build and dependency "
                        f"files are included in your `.bentoignore`, or try running Bento on specific files via `bento "
                        f"check [PATH]`",
                        extra=-4,
                    ),
                    err=True,
                    fg=bento.util.Colors.WARNING,
                    bold=False,
                )
                bento.util.echo_newline()
        else:
            logging.debug(f"Bento run completed in less than {SLOW_RUN_SECONDS} s.")

    def _setup_tool(
        self,
        bar: Optional[tqdm],
        ix: int,
        tool: Tool,
        max_bar_value: int,
        end_text: str,
    ) -> None:
        """
        Ensures that a tool is installed.

        :param bar: Any associated progress bar
        :param ix: The bar index
        :param tool: The tool
        """
        with self._updating_bar(
            bar, ix, 0, max_bar_value, bento.util.SETUP_TEXT, end_text
        ):
            tool.setup()

    def _run_single_tool(
        self, bar: Optional[tqdm], ix: int, tool: Tool, baseline: Baseline
    ) -> ToolResults:
        """
        Returns results for running a previously installed tool.

        :param bar: The progress bar associated with this tool
        :param ix: The current tool index
        :param tool: The tool itself
        :param paths: Paths to pass to the tool
        :param baseline: Any baseline to subtract from this tool
        :return: Tool results
        """
        with self._updating_bar(
            bar,
            ix,
            START_RUN_BAR_VALUE,
            DONE_BAR_VALUE,
            bento.util.PROGRESS_TEXT,
            bento.util.DONE_TEXT,
        ):
            results = bento.result.filtered(
                tool.tool_id(), tool.results(self.paths, self.use_cache), baseline
            )

        return results

    def _setup_and_run_single_tool(
        self, baseline: Baseline, index_and_tool: Tuple[int, Tool]
    ) -> RunResults:
        """Runs a tool and filters out existing findings using baseline"""

        ix, tool = index_and_tool
        end_of_setup_bar_value = (
            DONE_BAR_VALUE if self.install_only else START_RUN_BAR_VALUE
        )
        end_of_setup_text = (
            bento.util.DONE_TEXT if self.install_only else bento.util.PROGRESS_TEXT
        )
        bar = self._bars[ix] if self.show_bars else None

        try:
            before = time.time()
            logging.debug(f"{tool.tool_id()} start")

            if not self.skip_setup:
                self._setup_tool(
                    bar, ix, tool, end_of_setup_bar_value, end_of_setup_text
                )
            after_setup = time.time()

            results: ToolResults = []
            if not self.install_only:
                results = self._run_single_tool(bar, ix, tool, baseline)

            after = time.time()
            logging.debug(
                f"{tool.tool_id()} completed in {(after - before):2f} s (setup in {(after_setup - before):2f} s)"
            )

            return tool.tool_id(), results
        except Exception as e:
            logging.error(traceback.format_exc())
            return tool.tool_id(), e

    def parallel_results(
        self, tools: Iterable[Tool], baseline: Baseline, keep_bars: bool = True
    ) -> Collection[RunResults]:
        """Runs all tools in parallel.

        Each tool is optionally run against a list of files. For each tool, it's results are
        filtered to those results not appearing in the whitelist.

        A progress bar is emitted to stderr for each tool.

        Parameters:
            tools: (iterable): Tools to run
            baseline (set): The set of whitelisted finding hashes
            keep_bars (bool): If true, progress bars are preserved after run (default True)

        Returns:
            (collection): For each tool, a `RunResult`, which is a tuple of (`tool_id`, `findings`)
        """
        indices_and_tools = list(enumerate(tools))
        n_tools = len(indices_and_tools)

        if n_tools == 0:
            raise NoToolsConfiguredException()

        if self.show_bars:
            self._setup_bars(indices_and_tools)
        self._run = [True for _, _ in indices_and_tools]
        self._done = False

        slow_run_thread = threading.Thread(
            name="slow_run_timer", target=self._echo_slow_run
        )
        slow_run_thread.start()

        with ThreadPool(n_tools) as pool:
            # using partial to pass in multiple arguments to __tool_filter
            func = partial(Runner._setup_and_run_single_tool, self, baseline)
            all_results = pool.map(func, indices_and_tools)

        self._done = True
        slow_run_thread.join()

        if self.show_bars:
            if keep_bars:
                for _ in self._bars:
                    click.echo("", err=True)
            for b in self._bars:
                b.close()
            if keep_bars:
                # Progress bars terminate on whitespace
                bento.util.echo_newline()

        return all_results



import re
from pathlib import Path
from typing import Any, Dict, Iterable, Iterator, List, Pattern, Type

import yaml

from bento.constants import GREP_CONFIG_FILE_NAME
from bento.parser import Parser
from bento.tool import JsonR, output
from bento.violation import Violation


class GrepParser(Parser[JsonR]):
    def to_violation(self, output_rule: Dict[str, Any]) -> Violation:
        output = output_rule["output"]
        check_id = output_rule["id"]
        message = output_rule.get("message")
        parts = output.split(":")
        path = parts[0]
        path = self.trim_base(path)
        line_no = int(parts[1])
        code_snippet = ":".join(parts[2:])
        return Violation(
            tool_id=GrepTool.TOOL_ID,
            check_id=check_id,
            path=path,
            line=line_no,
            column=1,
            message=message or code_snippet,
            severity=2,
            syntactic_context=code_snippet or "<no context>",
        )

    def parse(self, input_json: JsonR) -> List[Violation]:
        return [self.to_violation(r) for r in input_json]


class GrepTool(output.Json):
    TOOL_ID = "grep"  # to-do: versioning?
    PROJECT_NAME = "Grep"

    @property
    def parser_type(self) -> Type[Parser]:
        return GrepParser

    @classmethod
    def tool_id(self) -> str:
        return GrepTool.TOOL_ID

    @classmethod
    def tool_desc(cls) -> str:
        return "Matches user-provided regex patterns (experimental)"

    @property
    def project_name(self) -> str:
        return GrepTool.PROJECT_NAME

    @property
    def file_name_filter(self) -> Pattern:
        return re.compile(r".*\b")

    def extra_cache_paths(self) -> List[Path]:
        return [(self.context.base_path / GREP_CONFIG_FILE_NAME).resolve()]

    def matches_project(self, files: Iterable[Path]) -> bool:
        # disabled by default for now
        return False

    def setup(self) -> None:
        pass  # TODO don't assume grep exists

    def query_rule(
        self, rule: Dict[str, Any], all_source_dirs: List[str]
    ) -> Iterator[str]:
        includes = []

        regex = rule.get("regex")
        if regex:
            file_ext_filters = rule.get("file_extentions", [])
            for file_ext_filter in file_ext_filters:
                includes.extend(["--include", file_ext_filter])
            cmd = ["grep"] + includes + ["-nRI", regex] + all_source_dirs
        yield self.execute(cmd, capture_output=True).stdout

    def run(self, files: Iterable[str]) -> JsonR:
        all_source_dirs = [f"{source_dir}" for source_dir in files]

        try:
            with (self.context.base_path / GREP_CONFIG_FILE_NAME).open() as grep_file:
                yml = yaml.safe_load(grep_file)
            grep_rules = yml.get("patterns", [])
        except FileNotFoundError:
            grep_rules = []

        all_output = [
            {"output": line, **rule}
            for rule in grep_rules
            for output in self.query_rule(rule, all_source_dirs)
            for line in output.split("\n")
            if line
        ]
        return all_output



import logging
import time
from typing import Collection, Iterable, Tuple

import click

from bento.constants import IGNORE_FILE_NAME
from bento.result import Baseline
from bento.target_file_manager import NoGitHeadException, TargetFileManager
from bento.tool import Tool
from bento.tool_runner import Runner, RunResults, RunStep
from bento.util import echo_warning


# TODO baseline removal should not be part of tool running
def orchestrate(
    baseline: Baseline,
    target_file_manager: TargetFileManager,
    staged: bool,
    tools: Iterable[Tool],
) -> Tuple[Collection[RunResults], float]:
    """
        Manages interactions between TargetFileManager, Runner and Tools

        Uses passed target_file_manager, staged flag, and tool list to setup
        Runner and runs tools on relevant files, returning aggregated output
        of tool running and time to run all tools in parallel
    """
    elapsed = 0.0
    if staged:
        head_baseline, elapsed = _calculate_head_comparison(target_file_manager, tools)
        for t in tools:
            tool_id = t.tool_id()
            if tool_id not in baseline:
                baseline[tool_id] = head_baseline.get(tool_id, set())
            else:
                baseline[tool_id].update(head_baseline.get(tool_id, set()))

    with target_file_manager.run_context(staged, RunStep.CHECK) as target_paths:
        use_cache = not staged  # if --all then can use cache
        skip_setup = staged  # if check --all then include setup
        runner = Runner(paths=target_paths, use_cache=use_cache, skip_setup=skip_setup)

        if len(runner.paths) == 0:
            echo_warning(
                f"Nothing to check or archive. Please confirm that changes are staged and not excluded by `{IGNORE_FILE_NAME}`. To check all Git tracked files, use `--all`."
            )
            click.secho("", err=True)
            all_results: Collection[RunResults] = []
            elapsed = 0.0
        else:
            before = time.time()
            all_results = runner.parallel_results(tools, baseline)
            elapsed += time.time() - before

    return all_results, elapsed


def _calculate_head_comparison(
    target_file_manager: TargetFileManager, tools: Iterable[Tool]
) -> Tuple[Baseline, float]:
    """
    Calculates a baseline consisting of all findings from the branch head

    If no HEAD branch exists return empty baseline

    :param paths: Which paths are being checked
    :param tools: Which tools to check
    :return: The branch head baseline
    """
    try:
        with target_file_manager.run_context(True, RunStep.BASELINE) as target_paths:
            runner = Runner(paths=target_paths, use_cache=True, skip_setup=True)
            if len(runner.paths) > 0:
                before = time.time()
                comparison_results = runner.parallel_results(tools, {}, keep_bars=False)
                baseline = {
                    tool_id: {f.syntactic_identifier_str() for f in findings}
                    for tool_id, findings in comparison_results
                    if isinstance(findings, list)
                }
                elapsed = time.time() - before
                return baseline, elapsed
            else:
                return {}, 0.0
    except NoGitHeadException:
        logging.debug("No git head found so defaulting to empty head baseline")
        return {}, 0.0



import os
from typing import Any, List

from bento.util import AutocompleteSuggestions


def list_paths(ctx: Any, args: List[str], incomplete: str) -> AutocompleteSuggestions:
    """
    Lists paths when tab autocompletion is used on a path argument.

    Note that click always adds a space at the end of a suggestion, so repeated tabbing
    can not be used to fill a path. :(

    :param ctx: Unused
    :param args: Unused
    :param incomplete: Any partial completion currently under the cursor
    :return: A list of completion suggestions
    """
    # Cases for "incomplete" variable:
    #   - '': Search '.', no filtering
    #   - 'part_of_file': Search '.', filter
    #   - 'path/to/dir/': Search 'path/to/dir', no filter
    #   - 'path/to/dir/part_of_file': Search 'path/to/dir', filter
    dir_root = os.path.dirname(incomplete)
    path_stub = incomplete[len(dir_root) :]
    if path_stub.startswith("/"):
        path_stub = path_stub[1:]
    if dir_root == "":
        dir_to_list = "."
    else:
        dir_to_list = dir_root
    return [
        os.path.join(dir_root, p)
        for p in os.listdir(dir_to_list)
        if not path_stub or p.startswith(path_stub)
    ]



import json
import logging
import os
import re
import shutil
import subprocess
from pathlib import Path
from typing import Any, Dict, Iterable, List, Pattern, Type

import yaml
from semantic_version import Version

import bento.constants as constants
from bento.parser import Parser
from bento.tool import JsonR, output, runner
from bento.tool.runner.js_tool import NpmDeps
from bento.violation import Violation

# Input example:
# [
#   {
#     "filePath":"/Users/nbrahms/dev/echelon-backend/r2c/bento/test/integration/init.js",
#     "messages":[
#       {"ruleId":"no-console","severity":1,"message":"Unexpected console statement.","line":1,"column":1,"nodeType":"MemberExpression","messageId":"unexpected","endLine":1,"endColumn":12},
#       {"ruleId":"semi","severity":2,"message":"Missing semicolon.","line":1,"column":15,"nodeType":"ExpressionStatement","fix":{"range":[14,14],"text":";"}}
#     ],
#     "errorCount":1,
#     "warningCount":1,
#     "fixableErrorCount":1,
#     "fixableWarningCount":0,
#     "source":"console.log(3)\n"
#   }
# ]
#

RC_ENVIRONMENTS = "env"
"""'env' field of .eslintrc.yml"""


class EslintParser(Parser[JsonR]):
    REACT_PREFIX = "react/"
    REACT_PREFIX_LEN = len(REACT_PREFIX)
    IMPORT_PREFIX = "import/"
    IMPORT_PREFIX_LEN = len(IMPORT_PREFIX)
    JSX_ALLY_PREFIX = "jsx-a11y/"
    JSX_ALLY_PREFIX_LEN = len(JSX_ALLY_PREFIX)
    TS_PREFIX = "@typescript-eslint/"
    TS_PREFIX_LEN = len(TS_PREFIX)

    @staticmethod
    def to_link(check_id: str) -> str:
        if check_id.startswith(EslintParser.REACT_PREFIX):
            trimmed = check_id[EslintParser.REACT_PREFIX_LEN :]
            return f"https://github.com/yannickcr/eslint-plugin-react/blob/master/docs/rules/{trimmed}.md"
        elif check_id.startswith(EslintParser.IMPORT_PREFIX):
            trimmed = check_id[EslintParser.IMPORT_PREFIX_LEN :]
            return f"https://github.com/benmosher/eslint-plugin-import/blob/master/docs/rules/{trimmed}.md"
        elif check_id.startswith(EslintParser.JSX_ALLY_PREFIX):
            trimmed = check_id[EslintParser.JSX_ALLY_PREFIX_LEN :]
            return f"https://github.com/evcohen/eslint-plugin-jsx-a11y/blob/master/docs/rules/{trimmed}.md"
        elif check_id.startswith(EslintParser.TS_PREFIX):
            trimmed = check_id[EslintParser.TS_PREFIX_LEN :]
            return f"https://github.com/typescript-eslint/typescript-eslint/blob/master/packages/eslint-plugin/docs/rules/{trimmed}.md"
        else:
            return f"https://eslint.org/docs/rules/{check_id}"

    def to_violation(
        self, result: Dict[str, Any], message: Dict[str, Any]
    ) -> Violation:
        path = self.trim_base(result["filePath"])
        startLine = message["line"]
        endLine = message.get("endLine", startLine)
        source = result["source"][startLine - 1 : endLine]  # line numbers are 1-indexed
        check_id = message.get("ruleId", None)
        if check_id:
            link = self.to_link(check_id)
        else:
            check_id = "error"
            link = ""

        return Violation(
            tool_id=EslintTool.ESLINT_TOOL_ID,
            check_id=check_id,
            path=path,
            line=startLine,
            column=message["column"],
            message=message["message"],
            severity=message["severity"],
            syntactic_context="\n".join(source).rstrip(),
            link=link,
        )

    def parse(self, tool_output: JsonR) -> List[Violation]:
        violations: List[Violation] = []
        for r in tool_output:
            r["source"] = r.get("source", "").split("\n")
            violations += [self.to_violation(r, m) for m in r["messages"]]
        return violations


class EslintTool(runner.Node, output.Json):
    ESLINT_TOOL_ID = "eslint"  # to-do: versioning?
    CONFIG_FILE_NAME = ".eslintrc.yml"
    PROJECT_NAME = "node-js"

    JS_NAME_PATTERN = re.compile(r".*\.(?:js|jsx|ts|tsx)\b")

    MANIFEST_PATH: Path = Path(__file__).parent.resolve() / "eslint"

    # Packages we always need no matter what.
    ALWAYS_NEEDED = {
        "eslint": Version("6.1.0"),
        "eslint-config-airbnb": Version("18.0.1"),
        "eslint-plugin-import": Version("2.18.2"),
        "eslint-plugin-jsx-a11y": Version("6.2.3"),
        "eslint-plugin-react": Version("7.14.3"),
        "eslint-plugin-react-hooks": Version("1.7.0"),
    }
    TYPESCRIPT_PACKAGES = {
        "@typescript-eslint/parser": Version("2.3.3"),
        "@typescript-eslint/eslint-plugin": Version("2.3.3"),
        "typescript": Version("3.6.4"),
    }

    # Never fire on 'window', etc.
    ALWAYS_INCLUDE_GLOBALS = ["browser", "commonjs", "es6", "node"]

    # Remaining environments are determined by inspecting package.json
    POSSIBLE_GLOBALS = [
        "applescript",
        "atomtest",
        "embertest",
        "greasemonkey",
        "jasmine",
        "jest",
        "jquery",
        "mango",
        "meteor",
        "mocha",
        "nashorn",
        "phantomjs",
        "prototypejs",
        "protractor",
        "qunit",
        "serviceworker",
        "shelljs",
        "webextensions",
    ]

    @property
    def parser_type(self) -> Type[Parser]:
        return EslintParser

    @classmethod
    def tool_id(self) -> str:
        return EslintTool.ESLINT_TOOL_ID

    @classmethod
    def tool_desc(cls) -> str:
        return "Identifies and reports on patterns in JavaScript and TypeScript"

    @property
    def install_location(self) -> Path:
        return self.base_path / constants.RESOURCE_PATH / "eslint"

    @property
    def project_name(self) -> str:
        project_deps = self._dependencies(location=self.base_path)
        all_used = [g for g in self.POSSIBLE_GLOBALS if g in project_deps]
        if self.__uses_typescript(project_deps):
            all_used.append("TypeScript")
        if self.__uses_react(project_deps):
            all_used.append("react")

        if all_used:
            fws = ", ".join(sorted(all_used))
            return f"{EslintTool.PROJECT_NAME} (with {fws})"
        return EslintTool.PROJECT_NAME

    @property
    def file_name_filter(self) -> Pattern:
        return self.JS_NAME_PATTERN

    @property
    def eslintrc_path(self) -> Path:
        return self.install_location / EslintTool.CONFIG_FILE_NAME

    def matches_project(self, files: Iterable[Path]) -> bool:
        return (self.context.base_path / "package.json").exists()

    def __uses_typescript(self, deps: NpmDeps) -> bool:
        # ts dependency shouldn't be in main deps, but, if it is, ok
        return "typescript" in deps

    def __uses_react(self, deps: NpmDeps) -> bool:
        return "react" in deps.main  # react dependency must be in main deps

    def __copy_eslintrc(self, identifier: str) -> None:
        logging.info(f"Using {identifier} .eslintrc configuration")
        shutil.copy(
            self.MANIFEST_PATH / f".eslintrc-{identifier}.yml", self.eslintrc_path
        )

    def _setup_env(self) -> None:
        self.install_location.mkdir(exist_ok=True, parents=True)
        eslint_package_path = self.install_location / "package.json"
        if not eslint_package_path.exists():
            logging.info("Creating eslint environment")
            shutil.copy(self.MANIFEST_PATH / "package.json", eslint_package_path)

    def __add_globals(self, deps: NpmDeps) -> None:
        """Adds global environments to eslintrc"""
        node_globals = [
            g for g in self.POSSIBLE_GLOBALS if g in deps
        ] + self.ALWAYS_INCLUDE_GLOBALS
        env_dict = dict((g, True) for g in node_globals)

        with self.eslintrc_path.open() as stream:
            rc = yaml.safe_load(stream)
        rc[RC_ENVIRONMENTS] = env_dict
        with self.eslintrc_path.open("w") as stream:
            yaml.safe_dump(rc, stream)

    def setup(self) -> None:
        self._setup_env()
        needed_packages: Dict[str, Version] = self.ALWAYS_NEEDED.copy()
        project_deps = self._dependencies(location=self.base_path)
        project_has_typescript = self.__uses_typescript(project_deps)
        project_has_react = self.__uses_react(project_deps)
        if project_has_typescript:
            needed_packages.update(self.TYPESCRIPT_PACKAGES)

        self._ensure_packages(needed_packages)
        self._ensure_node_version()

        # install .eslintrc.yml if necessary
        if not self.eslintrc_path.exists():
            logging.info(f"Installing {EslintTool.CONFIG_FILE_NAME}...")

            if project_has_react and project_has_typescript:
                self.__copy_eslintrc("react-and-typescript")
            elif project_has_react:
                self.__copy_eslintrc("react")
            elif project_has_typescript:
                self.__copy_eslintrc("typescript")
            else:
                self.__copy_eslintrc("default")

            self.__add_globals(project_deps)

    @staticmethod
    def raise_failure(cmd: List[str], result: subprocess.CompletedProcess) -> None:
        # Tool returned fialure, or did not return json
        raise subprocess.CalledProcessError(
            result.returncode, cmd, output=result.stdout, stderr=result.stderr
        )

    def run(self, files: Iterable[str]) -> JsonR:
        disables = [
            arg
            for d in self.config.get("ignore", [])
            for arg in ["--rule", f"{d}: off"]
        ]
        cmd = [
            "./node_modules/eslint/bin/eslint.js",
            "--no-eslintrc",
            "--no-ignore",
            "-c",
            str(self.eslintrc_path),
            "-f",
            "json",
            "--ext",
            "js,jsx,ts,tsx",
            "--ignore-pattern",
            ".bento/",
            "--ignore-pattern",
            "node_modules/",
        ] + disables
        for f in files:
            cmd.append(os.path.abspath(f))
        result = self.execute(
            cmd,
            cwd=self.install_location,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env={"TIMING": "1", **os.environ},
            check=False,
        )
        logging.debug(f"{self.tool_id()}: stderr:\n" + result.stderr[0:4000])
        logging.debug(f"{self.tool_id()}: stdout:\n" + result.stdout[0:4000])

        # Return codes:
        # 0 = no violations, 1 = violations, 2+ = tool failure
        if result.returncode > 1:
            self.raise_failure(cmd, result)

        try:
            # TODO: this double-parses, which we can avoid in the future by having type-parameterized parsers
            lines = result.stdout.split("\n")
            data = lines[0]
            timing = "\n".join(lines[1:])
            logging.debug(f"r2c.eslint: TIMING:\n{timing}")
            return json.loads(data.strip())
        except Exception as ex:
            logging.error("Could not parse json output of eslint tool", ex)
            self.raise_failure(cmd, result)
            return []  # Unreachable, but mypy is poor



import json
import re
from abc import abstractmethod
from pathlib import Path
from typing import Iterable, List, Mapping, Optional, Pattern

import attr
from semantic_version import SimpleSpec

from bento.parser import Parser
from bento.tool import JsonR, output, runner
from bento.util import fetch_line_in_file
from bento.violation import Violation


class BaseSgrepParser(Parser[JsonR]):
    @classmethod
    @abstractmethod
    def tool_id(cls) -> str:
        pass

    def parse(self, results: JsonR) -> List[Violation]:
        violations: List[Violation] = []
        for check in results:
            check_id = check["check_id"]
            path = self.trim_base(check["path"])
            start_line = check["start"]["line"]
            start_col = check["start"]["col"]
            # Custom way to get check_name for sgrep-lint:0.1.10
            message = check.get("extra", {}).get("message")
            source = (
                fetch_line_in_file(self.base_path / path, start_line)
                or "<no source found>"
            ).rstrip()
            violation = Violation(
                tool_id=self.tool_id(),
                check_id=check_id,
                path=path,
                line=start_line,
                column=start_col,
                message=message,
                severity=2,
                syntactic_context=source,
            )

            violations.append(violation)
        return violations


@attr.s
class BaseSgrepTool(runner.Docker, output.Json):
    DOCKER_IMAGE = "returntocorp/sgrep:0.4.8"
    FILE_NAME_FILTER = re.compile(r".*")
    BINARY_SPEC = SimpleSpec(">=0.4.8")

    @property
    def binary_executable(self) -> Optional[str]:
        return "semgrep"

    @property
    def allowed_binary_versions(self) -> SimpleSpec:
        return BaseSgrepTool.BINARY_SPEC

    @property
    @abstractmethod
    def config_str(self) -> str:
        """
        Returns the configuration argument to pass to sgrep
        """
        pass

    def get_config_path(self) -> Optional[Path]:
        """
        Returns the path to the sgrep configuration file _if it is a path_.
        Returns None otherwise.
        """
        return None

    @property
    def docker_image(self) -> str:
        return self.DOCKER_IMAGE

    @property
    def remote_code_path(self) -> str:
        return "/home/repo"

    @property
    def additional_file_targets(self) -> Mapping[Path, str]:
        config_path = self.get_config_path()
        if config_path:
            return {config_path: self.config_str}
        else:
            return {}

    @property
    def docker_command(self) -> List[str]:
        return ["--config", self.config_str, "--json"]

    @property
    def file_name_filter(self) -> Pattern:
        return self.FILE_NAME_FILTER

    @property
    def project_name(self) -> str:
        return "Python/JS"

    @classmethod
    def tool_desc(cls) -> str:
        return "Runs checks from r2c's check registry (experimental; requires Docker)"

    def run(self, files: Iterable[str]) -> JsonR:
        output_str = self.run_command(files).stdout.split("\n")
        stdout = output_str[-2]  # last line is blank
        output = json.loads(stdout)
        return output.get("results", [])



import logging
import os
import sys
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

import attr
import click
from packaging.version import InvalidVersion, Version

import bento.commands.autocomplete as autocomplete
import bento.constants as constants
import bento.content.register as content
import bento.decorators
import bento.extra
import bento.git
import bento.metrics
import bento.tool_runner
import bento.util
from bento.context import Context
from bento.error import InvalidVersionException, NonInteractiveTerminalException
from bento.network import post_metrics
from bento.util import echo_newline, persist_global_config, read_global_config

GLOBAL_GITIGNORE_PATTERN = ".bento/\n.bentoignore"


@attr.s(auto_attribs=True)
class Registrar:
    click_context: click.Context
    agree: bool
    is_first_run: bool = False
    context: Context = attr.ib(init=False)
    global_config: Dict[str, Any] = attr.ib(factory=read_global_config, init=False)
    email: Optional[str] = attr.ib()
    _displays_progress_bars: bool = attr.ib(init=False, default=False)

    @email.default
    def _get_email_from_environ(self) -> Optional[str]:
        return os.environ.get(constants.BENTO_EMAIL_VAR)

    def __attrs_post_init__(self) -> None:
        self.context = self.click_context.obj
        if self.global_config is None:
            self.is_first_run = True
            self.global_config = {}

    def _validate_interactivity(self) -> None:
        """
        Validates that this Bento session is running interactively

        :raises: SystemExit(3) if not interactive
        """
        is_interactive = sys.stdin.isatty() and sys.stderr.isatty()
        if not is_interactive:
            content.not_registered.echo()
            raise NonInteractiveTerminalException()

    def _show_welcome_message(self) -> None:
        """
        Displays a 'welcome to Bento' message

        Message is only displayed if registration is not skipped via command-line arguments

        :param agree: If the user has agreed to all prompts via the command line
        :param email: The user's email, if supplied via command line
        """
        if (
            self.email is None
            and "email" not in self.global_config
            or not self.agree
            and constants.TERMS_OF_SERVICE_KEY not in self.global_config
        ):
            content.welcome.echo()

    def _update_email(self) -> None:
        """
        Updates the user's global config with their email address

        If the user has passed an email on the command line, this logic is skipped.
        """
        # import inside def for performance
        from validate_email import validate_email

        if not self.email:
            self.email = self.global_config.get("email")

        if not self.email or not validate_email(self.email):
            content.UpdateEmail.leader.echo()

            email = None
            while not (email and validate_email(email)):
                self.context.start_user_timer()
                self._validate_interactivity()
                email = content.UpdateEmail.prompt.echo(
                    type=str, default=bento.git.user_email()
                )
                self.context.stop_user_timer()
                echo_newline()

            if email != constants.QA_TEST_EMAIL_ADDRESS:
                r = self._post_email_to_mailchimp(email)
                if not r:
                    content.UpdateEmail.failure.echo()

            self.global_config["email"] = email
            persist_global_config(self.global_config)

    @staticmethod
    def _post_email_to_mailchimp(email: str) -> bool:
        """
        Subscribes this email to the Bento mailing list

        :return: Mailchimp's response status
        """
        # import inside def for performance
        import requests

        r = requests.post(
            "https://waitlist.r2c.dev/subscribe", json={"email": email}, timeout=5
        )
        status: bool = r.status_code == requests.codes.ok
        data = [
            {
                "message": "Tried adding user to Bento waitlist",
                "user-email": email,
                "mailchimp_response": r.status_code,
                "success": status,
            }
        ]
        logging.info(f"Registering user with data {data}")
        post_metrics(data)
        return status

    def _confirm_tos_update(self) -> bool:
        """
        Interactive process to confirm updated agreement to the Terms of Service

        :return: If the user has agreed to the updated ToS
        """
        if constants.TERMS_OF_SERVICE_KEY not in self.global_config:
            content.ConfirmTos.fresh.echo()
        else:
            # We care that the user has agreed to the current terms of service
            tos_version = self.global_config[constants.TERMS_OF_SERVICE_KEY]

            try:
                agreed_to_version = Version(tos_version)
                if agreed_to_version == Version(constants.TERMS_OF_SERVICE_VERSION):
                    logging.info("User ToS agreement is current")
                    return True
            except InvalidVersion:
                content.ConfirmTos.invalid_version.echo()
                raise InvalidVersionException()

            content.ConfirmTos.upgrade.echo()

        self.context.start_user_timer()
        self._validate_interactivity()
        agreed = content.ConfirmTos.prompt.echo()
        echo_newline()
        self.context.stop_user_timer()

        if agreed:
            self.global_config[
                constants.TERMS_OF_SERVICE_KEY
            ] = constants.TERMS_OF_SERVICE_VERSION

            persist_global_config(self.global_config)
            return True
        else:
            content.ConfirmTos.error.echo()
            return False

    def _query_gitignore_update(self) -> Tuple[Path, bool]:
        """
        Determines if gitignore should be updated by init

        Requirements:
        - Interactive terminal
        - bento/ not in ignore file
        - User hasn't previously opted out
        - User agrees

        :return: A tuple of (the path to the global git ignore, whether to update the file)
        """

        gitignore_path = (
            bento.git.global_ignore_path(self.context.base_path)
            or constants.DEFAULT_GLOBAL_GIT_IGNORE_PATH
        )

        if sys.stderr.isatty() and sys.stdin.isatty():
            opt_out_value = self.global_config.get(constants.GLOBAL_GIT_IGNORE_OPT_OUT)
            user_opted_out = opt_out_value is not None and opt_out_value is True
            if user_opted_out:
                return gitignore_path, False
            has_ignore = None
            if gitignore_path.exists():
                with gitignore_path.open("r") as fd:
                    has_ignore = next(filter(lambda l: ".bento" in l, fd), None)
            if has_ignore is None:
                self.context.start_user_timer()
                if content.UpdateGitignore.confirm.echo(gitignore_path):
                    gitignore_path.parent.resolve().mkdir(exist_ok=True, parents=True)
                    content.UpdateGitignore.confirm_yes.echo()
                    return gitignore_path, True
                else:
                    self.global_config[constants.GLOBAL_GIT_IGNORE_OPT_OUT] = True

                    persist_global_config(self.global_config)
                    content.UpdateGitignore.confirm_no.echo()
                self.context.stop_user_timer()
        return gitignore_path, False

    def _update_gitignore_if_necessary(self, ignore_path: Path, update: bool) -> None:
        """
        Adds bento patterns to global git ignore if _query_gitignore_update returned a path
        """
        if update:
            self._displays_progress_bars = True
            on_done = content.UpdateGitignore.update.echo(ignore_path)
            bento.util.append_text_to_file(
                ignore_path,
                f"# Ignore bento tool run paths (this line added by `bento init`)\n{GLOBAL_GITIGNORE_PATTERN}",
            )
            on_done()
            logging.info(f"Added '.bento/' to {ignore_path}")

    def _suggest_autocomplete(self) -> None:
        """
        Suggests code to add to the user's shell config to set up autocompletion
        """
        shell = os.environ.get("SHELL")
        if not shell:
            return

        self._validate_interactivity()
        shell_cmd = shell.split("/")[-1]
        if shell_cmd in autocomplete.SUPPORTED:
            self.context.start_user_timer()
            should_add = content.SuggestAutocomplete.confirm.echo()
            self.context.stop_user_timer()
            if should_add:
                content.SuggestAutocomplete.confirm_yes.echo()
                self._displays_progress_bars = True
                on_done = content.SuggestAutocomplete.install.echo(
                    autocomplete.SUPPORTED[shell_cmd][0]
                )
                self.click_context.invoke(autocomplete.install_autocomplete, [False])
                on_done()
            else:
                content.SuggestAutocomplete.confirm_no.echo()

    def verify(self) -> bool:
        """
        Performs all necessary steps to ensure user registration:

        - Global config exists
        - User has agreed to Terms of Service
        - User has registered with email

        :param agree: If True, automatically confirms all yes/no prompts
        :param email: If exists, registers with this email
        :param context: The CLI context
        :return: Whether the user is properly registered after this function terminates
        """

        self._show_welcome_message()
        self._update_email()

        if not self.agree and not self._confirm_tos_update():
            return False

        # only ask about updating gitignore in init
        ignore_path: Optional[Path] = None
        update_ignore: Optional[bool] = None
        if not self.agree and self.context.is_init:
            ignore_path, update_ignore = self._query_gitignore_update()

        if self.is_first_run and not self.agree:
            self._suggest_autocomplete()

        if ignore_path is not None and update_ignore is not None:
            self._update_gitignore_if_necessary(ignore_path, update_ignore)

        if self._displays_progress_bars:
            content.finalize.echo()

        return True



import binascii
import textwrap
from typing import Any, Dict, Optional

import attr
import pymmh3 as mmh3


@attr.s(frozen=True, hash=False)
class Violation:
    """
    N.B.: line and column are 1-based, not 0-based
    """

    BASELINE_IGNORED_ITEMS = ["line", "column", "link", "filtered"]

    tool_id = attr.ib(type=str)
    check_id = attr.ib(type=str)
    path = attr.ib(type=str)
    # cmp is deprecated, but we need to use it for compatibility with 18.x.
    line = attr.ib(type=int, hash=None, cmp=False)
    column = attr.ib(type=int, hash=None, cmp=False)
    message = attr.ib(type=str, hash=None, cmp=False)
    severity = attr.ib(type=int, hash=None, cmp=False)
    syntactic_context = attr.ib(type=str, converter=textwrap.dedent)
    semantic_context = None
    filtered = attr.ib(type=Optional[bool], default=None, hash=None, cmp=False)
    link = attr.ib(type=Optional[str], default=None, hash=None, cmp=False, kw_only=True)

    def syntactic_identifier_int(self) -> int:
        # Use murmur3 hash to minimize collisions

        str_id = str((self.check_id, self.path, self.syntactic_context))
        return mmh3.hash128(str_id)

    def syntactic_identifier_str(self) -> str:
        id_bytes = int.to_bytes(
            self.syntactic_identifier_int(), byteorder="big", length=16, signed=False
        )
        return str(binascii.hexlify(id_bytes), "ascii")

    def __hash__(self) -> int:
        # attr.s equality uses all elements of syntactic_identifier, so
        # hash->equality contract is guaranteed
        return self.syntactic_identifier_int()

    def to_dict(self) -> Dict[str, Any]:
        d = attr.asdict(self)
        for i in Violation.BASELINE_IGNORED_ITEMS:
            d.pop(i)
        return d



import json
import re
import sys
from pathlib import PurePath
from typing import Any, Dict, Iterable, List, Optional, Pattern, Type

from semantic_version import SimpleSpec

from bento.parser import Parser
from bento.tool import JsonR, output, runner
from bento.util import fetch_line_in_file
from bento.violation import Violation

REMOTE_BASE_PATH = PurePath("/mnt")


class GosecParser(Parser[JsonR]):
    SEVERITIES = {"HIGH": 2, "MEDIUM": 1, "LOW": 0}

    def to_violation(self, result: Dict[str, Any]) -> Violation:
        start_line = int(result["line"])
        column = int(result["column"])
        check_id = result["rule_id"]
        message = result["details"]
        path = result["file"]

        level = result["severity"]
        severity = self.SEVERITIES.get(level, 0)

        link = result.get("cwe", {}).get("URL", "")

        line_of_code = (
            fetch_line_in_file(self.base_path / path, start_line) or "<no source found>"
        )

        return Violation(
            tool_id=GosecTool.TOOL_ID,
            check_id=check_id,
            path=path,
            line=start_line,
            column=column,
            message=message,
            severity=severity,
            syntactic_context=line_of_code,
            link=link,
        )

    def parse(self, results: JsonR) -> List[Violation]:
        return [self.to_violation(r) for r in results]


class GosecTool(runner.Docker, output.Json):
    """
    Runs securego/gosec.

    Note that the runtime of this tool is independent of the number of paths Bento is running on,
    and is always linear in the size of the project, due to limitations of the wrapped gosec interface.
    """

    TOOL_ID = "gosec"
    DOCKER_IMAGE = "securego/gosec:v2.2.0"
    FILE_FILTER = re.compile(r".*\.go$")
    BINARY_SPEC = SimpleSpec("~=2.2.0")
    VERSION_MATCH = re.compile(r"Version: ([\d.]+).*")

    @property
    def binary_executable(self) -> Optional[str]:
        return "gosec"

    @property
    def binary_version_pattern(self) -> Pattern:
        return self.VERSION_MATCH

    @property
    def allowed_binary_versions(self) -> SimpleSpec:
        return self.BINARY_SPEC

    @property
    def parser_type(self) -> Type[Parser]:
        return GosecParser

    @classmethod
    def tool_id(cls) -> str:
        return cls.TOOL_ID

    @classmethod
    def tool_desc(cls) -> str:
        return "Finds security bugs in Go code"

    @property
    def file_name_filter(self) -> Pattern:
        return self.FILE_FILTER

    @property
    def project_name(self) -> str:
        return "Go"

    @property
    def docker_image(self) -> str:
        return self.DOCKER_IMAGE

    @property
    def docker_command(self) -> List[str]:
        return ["--fmt", "json"]

    @property
    def remote_code_path(self) -> str:
        return str(self.base_path) if self.use_binary else str(REMOTE_BASE_PATH)

    def is_allowed_returncode(self, returncode: int) -> bool:
        return returncode == 0 or returncode == 1

    # gosec only operates on modules, not individual files, so we
    # set the max batch size to effectively unbounded, and filter the results
    # returned by gosec
    #
    # Unfortunately this means the runtime of gosec is constant with the number
    # of queried paths, and linear in the project size :(
    @classmethod
    def max_batch_size(cls) -> int:
        return sys.maxsize

    def assemble_full_command(self, targets: Iterable[str]) -> List[str]:
        return self.docker_command + [self.remote_code_path]

    def filter_result_paths(self, results: JsonR, files: Iterable[str]) -> JsonR:
        """Filters gosec results to only files that we care about"""

        def strip_base(r: Dict[str, Any]) -> Dict[str, Any]:
            r["file"] = str(PurePath(r["file"]).relative_to(self.remote_code_path))
            return r

        to_keep = {str(PurePath(f).relative_to(self.base_path)) for f in files}
        stripped = (strip_base(r) for r in results)
        out = [r for r in stripped if r["file"] in to_keep]
        return out

    def run(self, files: Iterable[str]) -> JsonR:
        output = self.run_command(files).stdout
        all_results = json.loads(output).get("Issues", [])
        return self.filter_result_paths(all_results, files)



from pathlib import Path
from typing import List, Optional

from bento.constants import SUPPORT_EMAIL_ADDRESS


class BentoException(Exception):
    """
    Parent class of all exceptions we anticipate in Bento commands.

    All BentoExceptions are caught and their error messages are printed to the console.
    Inherit from this class to override the message and error code. Make sure to call super().__init__() in the child constructor
    """

    def __init__(self) -> None:
        self.msg: Optional[str] = None
        self.code = 3


class OutdatedPythonException(BentoException):
    def __init__(self) -> None:
        super().__init__()
        self.msg = "Bento requires Python 3.6+. Please ensure you have Python 3.6+ and installed Bento via `pip3 install bento-cli`."


class InvalidRegistrationException(BentoException):
    def __init__(self) -> None:
        super().__init__()
        self.msg = "Could not verify the user's registration."


class NoConfigurationException(BentoException):
    def __init__(self) -> None:
        super().__init__()
        self.msg = "No Bento configuration found. Please run `bento init`."


class NoIgnoreFileException(BentoException):
    def __init__(self, path: Path) -> None:
        super().__init__()
        self.msg = f"No ignore file found (looked in {path}). Please run `bento init`."


class NotAGitRepoException(BentoException):
    def __init__(self) -> None:
        super().__init__()
        self.msg = "Not a git repository."


class ExistingGitHookException(BentoException):
    def __init__(self, hook_path: str) -> None:
        super().__init__()
        self.msg = f"Autorun could not be configured: A legacy pre-commit hook exists. Please remove {hook_path}.pre-bento to continue."


class ToolRunException(BentoException):
    def __init__(self, message: str = "") -> None:
        super().__init__()
        self.msg = f"Bento failed while running a tool: {message}"


class NonInteractiveTerminalException(BentoException):
    def __init__(self) -> None:
        super().__init__()
        self.msg = "Bento can't run this command in a non-interactive terminal."


class InvalidVersionException(BentoException):
    def __init__(self) -> None:
        super().__init__()
        self.msg = (
            "Bento detected an invalid version in your global configuration file."
        )


class InvalidToolException(BentoException):
    def __init__(self, tool: str, all_tools: str) -> None:
        super().__init__()
        self.msg = f"No tool named '{tool}'. Configured tools are {all_tools}"


class EnabledToolNotFoundException(BentoException):
    def __init__(self, tool: str) -> None:
        super().__init__()
        self.msg = (
            f"Bento didn't recognize the tool named '{tool}', "
            "even though it's enabled in your project configuration.\n"
            "Try updating Bento with `pip install --upgrade bento-cli`."
        )


class UnsupportedGitStateException(BentoException):
    def __init__(self, extra: str = "") -> None:
        super().__init__()
        self.msg = f"Cannot complete running command in this git state. {extra}"


class DockerFailureException(BentoException):
    def __init__(self) -> None:
        super().__init__()
        self.msg = "Failed to run docker. Please confirm docker is installed and its daemon is running in user mode."


class NoToolsConfiguredException(BentoException):
    def __init__(self) -> None:
        super().__init__()
        self.msg = f"Please specify a tool with `--tool TOOL_NAME`."


class UnsupportedCIProviderException(BentoException):
    def __init__(self) -> None:
        super().__init__()
        self.msg = (
            "Automated CI installation is avaiable only for GitHub Actions currently, "
            "but your project does not seem to be using GitHub.\n"
            f"Please contact {SUPPORT_EMAIL_ADDRESS} if this is incorrect."
        )


class NodeError(Exception):
    """
    Node not found or node version not supported by ESLint 6
    """


class IncompatibleParametersException(BentoException):
    def __init__(self, parameters: List[str]) -> None:
        super().__init__()
        self.msg = f"Parameters {parameters} can not be used together."


class MultipleErrorsException(BentoException):
    def __init__(self, errors: List[BentoException]) -> None:
        super().__init__()
        messages = "\n\n".join(
            error.msg if error.msg else str(error) for error in errors
        )
        self.msg = f"The following errors occurred during execution:\n\n{messages}"



from bento.tool.tool import JsonR, Tool

"""
Classes that define the output format of a tool
"""

Str = Tool[str]
"""Output is a raw string"""

Json = Tool[JsonR]
"""Output is parsed JSON (of type bento.tool.JsonR)"""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bento',
 'bento.commands',
 'bento.content',
 'bento.extra',
 'bento.formatter',
 'bento.tool',
 'bento.tool.runner']

package_data = \
{'': ['*'], 'bento': ['configs/*'], 'bento.extra': ['eslint/*']}

install_requires = \
['PyYAML>=5.1.2',
 'attrs>=18.2.0,<=19.3.0',
 'click>=7.0,<8.0',
 'docker>=3.7,<4.0',
 'frozendict>=1.2,<2.0',
 'gitpython>=2.1,<3.0',
 'packaging>=14.0',
 'pre-commit>=1.0.0,<=1.18.3',
 'psutil>=5.6.3,<5.7.0',
 'pymmh3>=0.0.5,<0.1.0',
 'semantic-version>=2.8.0,<2.9.0',
 'tqdm>=4.36.1,<4.37.0',
 'validate-email>=1.3,<2.0']

entry_points = \
{'console_scripts': ['bentoh = bento.__main__:headless']}

setup_kwargs = {
    'name': 'bento-headless',
    'version': '0.11.2',
    'description': 'Git-aware utility for automated program analysis',
    'long_description': '<h3 align="center">\n  [alpha] A Git-aware CLI for running semgrep patterns in the developer and CI workflow.\n</h3>\n\n<p align="center">\n  <a href="#installation">Installation</a>\n  <span> · </span>\n  <a href="#usage">Usage</a>\n  <span> · </span>\n  <a href="#help-and-community">Help & Community</a>\n</p>\n\n<p align="center">\n  <a href="https://pypi.org/project/bento-cli/">\n    <img alt="PyPI" src="https://img.shields.io/pypi/v/bento-cli?style=flat-square&color=blue">\n  </a>\n  <a href="https://pypi.org/project/bento-cli/">\n    <img alt="PyPI - Downloads" src="https://img.shields.io/pypi/dm/bento-cli?style=flat-square&color=green">\n  </a>\n  <a href="https://github.com/returntocorp/bento/issues/new/choose">\n    <img src="https://img.shields.io/badge/issues-welcome-green?style=flat-square" alt="Issues welcome!" />\n  </a>\n  <a href="https://twitter.com/intent/follow?screen_name=r2cdev">\n    <img src="https://img.shields.io/twitter/follow/r2cdev?label=Follow%20r2cdev&style=social&color=blue" alt="Follow @r2cdev" />\n  </a>\n</p>\n\n## Installation\n\nRequires [Python 3.6+](https://www.python.org/downloads/) and [Docker 19.03+](https://docs.docker.com/get-docker/). It runs on macOS and Linux.\n\nIn a Git project directory:\n\n```bash\n$ pip3 install bento-headless\n```\n\n## Usage\n\n### Upgrading\n\n```bash\n$ pip3 install --upgrade bento-headless\n```\n\n### Command line options\n\n```\n$ bentoh --help\nUsage: bentoh [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  -h, --help  Show this message and exit.\n  --version   Show the version and exit.\n\nCommands:\n  archive  Suppress current findings.\n  check    Checks for new findings.\n\n  To get help for a specific command, run `bentoh COMMAND --help`\n```\n\n### Run custom `semgrep` checks on staged diffs\n\nSee [semgrep Configuration](https://github.com/returntocorp/semgrep/blob/develop/docs/config/advanced.md) for how to write custom rule files\n\n```\nvi .bento/semgrep.yml\nbentoh check\n```\n\n### Format output as JSON\n\n```\nbentoh check -f json\n```\n\n### Run on file system current state\n\n```\nbentoh check --all\n```\n\n### Run on staged diffs in a directory\n\n```\nbentoh check src\n```\n\n### Ignore current findings\n\n```\nbentoh archive\n```\n\n### Run public semgrep checks on staged diffs\n\n```\nBENTO_REGISTRY=r/r2c.python bentoh check\n```\n\n### Run checks from extensions\n\n```\nbentoh check -t gosec -t r2c.registry.latest\n```\n\n### Exit codes\n\n`bentoh check` may exit with the following exit codes:\n\n- `0`: Bento ran successfully and found no errors\n- `2`: Bento ran successfully and found issues in your code\n- `3`: Bento or one of its underlying tools failed to run\n\n## Extensions\n\n`bentoh` ships with the following extensions:\n\n| Extension           | Description                                                                           |\n| ------------------- | ------------------------------------------------------------------------------------- |\n| bandit              | Finds common security issues in Python code                                           |\n| dlint               | A tool for encouraging best coding practices and helping ensure Python code is secure |\n| eslint              | Identifies and reports on patterns in JavaScript and TypeScript                       |\n| flake8              | Finds common bugs in Python code                                                      |\n| gosec               | Finds security bugs in Go code                                                        |\n| hadolint            | Finds bugs in Docker files (requires Docker)                                          |\n| r2c.boto3           | Checks for the AWS boto3 library in Python                                            |\n| r2c.flask           | Checks for the Python Flask framework                                                 |\n| r2c.jinja           | Finds common security issues in Jinja templates                                       |\n| r2c.registry.latest | Runs checks from r2c\'s check registry (experimental; requires Docker)                 |\n| r2c.requests        | Checks for the Python Requests framework                                              |\n| shellcheck          | Finds bugs in shell scripts (requires Docker)                                         |\n\n## Help and community\n\nNeed help or want to share feedback? We’d love to hear from you!\n\n- Email us at [support@r2c.dev](mailto:support@r2c.dev)\n- Join #general in our [community Slack](https://join.slack.com/t/r2c-community/shared_invite/enQtNjU0NDYzMjAwODY4LWE3NTg1MGNhYTAwMzk5ZGRhMjQ2MzVhNGJiZjI1ZWQ0NjQ2YWI4ZGY3OGViMGJjNzA4ODQ3MjEzOWExNjZlNTA)\n- [File an issue](https://github.com/returntocorp/bento/issues/new?assignees=&labels=bug&template=bug_report.md&title=) or [submit a feature request](https://github.com/returntocorp/bento/issues/new?assignees=&labels=feature-request&template=feature_request.md&title=) directly on GitHub\n\nWe’re constantly shipping new features and improvements.\n\n## License and legal\n\nPlease refer to the [terms and privacy document](https://github.com/returntocorp/bento/blob/master/PRIVACY.md).\n\n</br>\n</br>\n<p align="center">\n    <img src="https://web-assets.r2c.dev/r2c-logo-silhouette.png?gh" height="24" alt="r2c logo"/>\n</p>\n<p align="center">\n    Copyright (c) <a href="https://r2c.dev">r2c</a>.\n</p>\n',
    'author': 'Return To Corporation',
    'author_email': 'hello@r2c.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://bento.dev',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import sys
from pathlib import Path

import click
import yaml

import bento.content.ci as content
import bento.git
import bento.tool_runner
from bento.context import Context
from bento.decorators import with_metrics
from bento.error import (
    InvalidRegistrationException,
    NotAGitRepoException,
    UnsupportedCIProviderException,
)
from bento.util import echo_next_step, echo_success, echo_warning, read_global_config


def is_ci_configured(context: Context) -> bool:
    return context.gh_actions_file_path.exists()


def is_ci_provider_supported(repo_path: Path) -> bool:
    origin_url = bento.git.url(repo_path)
    if not origin_url:
        return False
    return "github" in origin_url


def _raise_if_unsupported(repo_path: Path) -> None:
    repo = bento.git.repo(repo_path)
    if repo is None:
        raise NotAGitRepoException()

    if not is_ci_provider_supported(repo_path):
        raise UnsupportedCIProviderException()


def _get_user_email() -> str:
    # the Registrar already ran and ensured we have an email
    global_config = read_global_config()
    if global_config is None:  # handling this case only because of mypy
        raise InvalidRegistrationException()
    return global_config["email"]


def _write_gh_actions_config(target_path: Path, email: str) -> None:
    source_path = Path(__file__).parent.parent / "configs" / "gh-actions.yml"

    target_path.parent.mkdir(parents=True, exist_ok=True)

    with source_path.open() as source_file, target_path.open("w") as target_file:
        gh_actions_config = yaml.load(source_file, Loader=yaml.SafeLoader)
        job_config = gh_actions_config["jobs"]["bento"]
        job_config["steps"][-1]["with"]["acceptTermsWithEmail"] = email
        yaml.dump(gh_actions_config, target_file)


def _delete_gh_actions_config(*, path: Path, root_path: Path) -> None:
    """Removes the Bento GitHub Actions config file.

    Also cleans up parent directories recursively if they're left empty.
    """
    path.unlink()

    current_dir = path.parent
    while current_dir != root_path:
        try:
            current_dir.rmdir()
        except OSError:
            return  # directory not empty, cleanup done
        current_dir = current_dir.parent


@click.command(name="ci")
@click.pass_obj
@with_metrics
def install_ci(context: Context) -> None:
    """
    Configures Bento to run in CI.
    """
    _raise_if_unsupported(context.base_path)

    pretty_path = context.pretty_path(context.gh_actions_file_path)
    config_directory = pretty_path.parts[0]

    if not context.is_init:
        # confirmation was handled as part of init
        content.Install.banner.echo()

        if is_ci_configured(context):
            content.Overwrite.warn.echo(str(pretty_path).strip())

            is_overwrite_confirmed = content.Overwrite.confirm.echo()
            content.Overwrite.after_confirm.echo()
            if not is_overwrite_confirmed:
                sys.exit()

        on_done = content.Install.progress.echo(config_directory)

    _write_gh_actions_config(context.gh_actions_file_path, _get_user_email())

    if not context.is_init:
        on_done()

        content.Install.after_progress.echo()
        content.Install.finalize_ci.echo()


@click.command(name="ci")
@click.pass_obj
@with_metrics
def uninstall_ci(context: Context) -> None:
    """
    Configures Bento to NOT run in CI.
    """
    repo = bento.git.repo(context.base_path)
    if repo is None:
        raise NotAGitRepoException()

    if not context.gh_actions_file_path.exists():
        echo_warning(
            "Not uninstalling CI config: Bento is not configured for CI on this project."
        )
        sys.exit(1)

    _delete_gh_actions_config(
        path=context.gh_actions_file_path, root_path=context.base_path
    )

    echo_success("Uninstalled Bento from CI.")
    echo_next_step("To re-enable CI integration", "bento enable ci")



import sys
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Generic, List, Tuple, TypeVar, Union, cast

import attr
from click import confirm, prompt, secho, style

import bento.util

# Return type for renderer
R = TypeVar("R", covariant=True)

# Represents a text transformation
Processor = Callable[[str], str]


@attr.s(auto_attribs=True)
class Link:
    """
    Represents a link definition

    :param anchor: Link anchor text
    :param href: Link target
    """

    anchor: str
    href: str


class Processors:
    identity: Processor = lambda text: text

    @staticmethod
    def ljust(extra: int = 0, char: str = " ") -> Processor:
        """
        Left-justifies text, filling width with specified character

        :param extra: Number of extra characters to add to justification (useful for
                      accommodating ANSI control characters
        :param char: The fill character
        """
        return lambda text: text.ljust(bento.util.PRINT_WIDTH + extra, char)

    @staticmethod
    def wrap(extra: int = 0) -> Processor:
        """
        Wraps text to print width

        :param extra: Number of extra characters to add to margin (useful for
                      accommodating ANSI control characters
        """
        return lambda text: bento.util.wrap(text, extra)

    @staticmethod
    def wrap_link(links: List[Link], extra: int = 0, **kwargs: Any) -> Processor:
        """
        Per wrap, except links are rendered without affecting the wrap width.

        In addition to the specification of process_wrap, a sequence of "link" values controls link processing.

        Links are only rendered if the terminal is a known supporting terminal for OSC8 links. If not,
        links are rendered as their anchor text only.

        :param links: Link definitions
        :param extra: Any extra width to apply
        :param kwargs: Styling rules passed to text
        """

        def _wrap_link(text: str) -> str:
            wrapped = bento.util.wrap(text, extra)

            def find_loc(anchor: str) -> Tuple[int, str]:
                """
                Finds the position of the anchor string in the wrapped text

                Note that the anchor string itself may be wrapped, so we return
                both the position, and the value of the (possibly wrapped) anchor string.
                """
                pos = wrapped.find(anchor)
                if pos < 0:
                    # Text was likely wrapped
                    anchor_it = [
                        f"{anchor[:ix].rstrip()}\n{anchor[ix:]}"
                        for ix in range(len(anchor))
                    ]
                    pos_it = ((wrapped.find(a), a) for a in anchor_it)
                    pos, anchor = next(
                        ((p, a) for p, a in pos_it if p > 0), (-1, anchor)
                    )
                    if pos < 0:
                        raise ValueError(f"'{anchor}' does not appear in '{text}'")
                return pos, anchor

            with_locs = sorted(
                [(find_loc(link.anchor), link.href) for link in links],
                key=(lambda t: t[0][0]),
            )
            out = ""
            current = 0
            for loc_anchor, href in with_locs:
                loc, anchor = loc_anchor
                out += style(wrapped[current:loc], **kwargs)
                out += bento.util.render_link(
                    anchor, href, print_alternative=False, pipe=sys.stderr
                )
                current = loc + len(anchor)
            out += style(wrapped[current:], **kwargs)

            return out

        return _wrap_link


class Content(ABC):
    """
    Represents styled, processed, textual content

    Each implementation of this class should define the following keyword arguments:
    :param processor: The text processor to apply
    :param style: Style rules to apply
    """

    def __init__(self) -> None:
        self.processor = Processors.identity
        self.style: Dict[str, Any] = {}

    @abstractmethod
    def make(self, *args: Any) -> str:
        pass

    def expand(self, *args: Any, apply_style: bool = False) -> str:
        text = self.processor(self.make(*args))
        if apply_style:
            return style(text, **self.style)
        else:
            return text


@attr.s()
class Sub(Content):
    """
    Represents text taken from an argument

    Substitution uses the argument's `str` value.

    :param index: The index of the argument list to substitute
    """

    index = attr.ib(type=int)
    style = attr.ib(type=Dict[str, Any], kw_only=True, default={})
    processor = attr.ib(
        type=Callable[[str], str], kw_only=True, default=Processors.identity
    )

    def make(self, *args: Any) -> str:
        return str(args[self.index])


@attr.s()
class Text(Content):
    """
    Represents literal text

    :param text: The text
    """

    text = attr.ib(type=str)
    style = attr.ib(type=Dict[str, Any], kw_only=True, default={})
    processor = attr.ib(
        type=Callable[[str], str], kw_only=True, default=Processors.identity
    )

    def make(self, *args: Any) -> str:
        return self.text


@attr.s()
class Multi(Content):
    """
    Represents a concatenation of other content

    :param items: The wrapped content; a literal string may also be used instead of unstyled, unprocessed Text content
    """

    items = attr.ib(type=List[Union[str, Content]])
    style = attr.ib(type=Dict[str, Any], kw_only=True, default={})
    processor = attr.ib(
        type=Callable[[str], str], kw_only=True, default=Processors.identity
    )

    def make(self, *args: Any) -> str:
        converted = (c if isinstance(c, Content) else Text(c) for c in self.items)
        parts = (c.expand(*args, apply_style=True) for c in converted)
        return "".join(parts)


@attr.s
class Renderer(ABC, Generic[R]):
    """
    Renders content from a YAML file

    To create, use:

      [RendererType](content, ...)

    To use, invoke:

      renderer.echo(*args, **kwargs)

    `args` are a list of arguments to be used in substitution content;
    `kwargs` are passed directly to the renderer, and their allowed values depend
    on the renderer in question.

    :param content: A Content object, defaults to empty text
    :param styling: A styling dictionary, defaults to no styling
    """

    content = attr.ib(type=Union[str, Content], default=Text(""))
    styling = attr.ib(type=Dict[str, Any], init=False)

    def __attrs_post_init__(self) -> None:
        if isinstance(self.content, str):
            self.content = Text(self.content)
        self.styling = self.content.style

    @abstractmethod
    def render(self, text: str, **kwargs: Any) -> R:
        """
        Renders content

        :param text: The styled and processed content
        :param kwargs: Any keyword arguments to be used by this renderer
        :return: This renderer's return value
        """
        pass

    def text(self, *args: Any) -> str:
        """
        Returns un-rendered (but styled) content text

        :param args: Arguments for substitution content
        """
        return cast(Content, self.content).expand(*args, apply_style=True)

    def echo(self, *args: Any, **kwargs: Any) -> R:
        """
        Renders content

        :param args: Arguments for substitution content
        :param kwargs: Renderer-specific arguments (see help for that renderer's `render` method)
        """
        text = cast(Content, self.content).expand(*args, apply_style=False)
        return self.render(text, **kwargs)


@attr.s
class Box(Renderer[None]):
    """
    Renders content in a Unicode-drawn box to stderr
    """

    def render(self, text: str, **kwargs: Any) -> None:
        bento.util.echo_box(text)


@attr.s(auto_attribs=True)
class Confirm(Renderer[bool]):
    """
    Renders a confirmation prompt, and returns confirmation state.

    :param options: Arguments to be passed to click.confirm
    """

    options: Dict[str, Any] = {}

    def render(self, text: str, **kwargs: Any) -> bool:
        """
        :param kwargs: Arguments passed to click.confirm (in addition to self.options)
        :return: The entered value
        """
        options = {**self.options, **kwargs}
        return confirm(style(text, **self.styling), err=True, **options)


@attr.s(auto_attribs=True)
class Prompt(Renderer[str]):
    """
    Renders a text prompt, and returns text state.

    :param options: Arguments to be passed to click.prompt
    """

    options: Dict[str, Any] = {}

    def render(self, text: str, **kwargs: Any) -> str:
        """
        :param kwargs: Arguments passed to click.prompt (in addition to self.options)
        :return: The entered value
        """
        options = {**self.options, **kwargs}
        return prompt(style(text, **self.styling), err=True, **options)


@attr.s(auto_attribs=True)
class Echo(Renderer[None]):
    """
    Renders content, appended together, to stderr

    :param newline: Whether to render a newline at the end of content (default True)
    """

    newline: bool = True

    def render(self, text: str, **kwargs: Any) -> None:
        secho(text, err=True, nl=self.newline, **self.styling)


@attr.s(auto_attribs=True)
class Error(Renderer[None]):
    """
    Renders error content to stderr
    """

    def render(self, text: str, **kwargs: Any) -> None:
        bento.util.echo_error(text)


@attr.s(auto_attribs=True)
class Newline(Renderer[None]):
    """
    Renders a newline

    :raises ValueError: If content is not empty text
    """

    def __attrs_post_init__(self) -> None:
        if not isinstance(self.content, Text) or self.content.text:
            raise ValueError("Newline renderer should not have content")

    def render(self, text: str, **kwargs: Any) -> None:
        bento.util.echo_newline()


@attr.s(auto_attribs=True)
class Success(Renderer[None]):
    """
    Renders success content to stderr
    """

    def render(self, text: str, **kwargs: Any) -> None:
        bento.util.echo_success(text)


@attr.s(auto_attribs=True)
class Warn(Renderer[None]):
    """
    Renders warning content to stderr
    """

    def render(self, text: str, **kwargs: Any) -> None:
        bento.util.echo_warning(text)


@attr.s(auto_attribs=True)
class Progress(Renderer[Callable[[], None]]):
    """
    Renders a progress bar

    For more information, see bento.util.echo_progress.

    :param extra: Extra width to pass to `echo_progress`
    :return: The progress done callback
    """

    extra: int = 0

    def render(self, text: str, **kwargs: Any) -> Callable[[], None]:
        return bento.util.echo_progress(text, self.extra, **kwargs)


class Steps:
    """
    Successively renders multiple Renderer objects
    """

    def __init__(self, *steps: Renderer) -> None:
        self.steps = steps

    def echo(self, *args: Any, **kwargs: Any) -> List[Any]:
        return [s.echo(*args, **kwargs) for s in self.steps]



from bento.renderer import (
    Box,
    Confirm,
    Echo,
    Error,
    Link,
    Multi,
    Newline,
    Processors,
    Progress,
    Prompt,
    Steps,
    Sub,
    Text,
    Warn,
)

not_registered = Steps(
    Error("This installation of Bento is not registered."),
    Echo(
        """
Please either:
◦ Register Bento by running it in an interactive terminal
◦ Run Bento with `--agree --email [EMAIL]`"""
    ),
)

welcome = Steps(
    Box("Global Bento Configuration"),
    Echo(
        Text(
            "Thanks for installing Bento, a free and opinionated toolkit for gradually adopting linters and program "
            "analysis in your codebase!",
            processor=Processors.wrap(),
        )
    ),
    Newline(),
)


finalize = Newline()


class UpdateEmail:
    leader = Steps(
        Echo(
            Multi(
                [
                    "Registration: ",
                    Text(
                        "We’ll use your email to provide support and share product updates. You can unsubscribe at "
                        "any time.",
                        style={"dim": True},
                    ),
                ],
                processor=Processors.wrap(extra=8),
            )
        ),
        Newline(),
    )
    prompt = Prompt("What is your email address?")
    failure = Steps(
        Newline(),
        Warn(
            "We were unable to subscribe you to the Bento mailing list (which means you may miss out on "
            "announcements!). Bento will continue running. Please shoot us a note via support@r2c.dev to debug. "
        ),
    )


class ConfirmTos:
    fresh = Steps(
        Echo(
            Multi(
                [
                    "Privacy: ",
                    Text(
                        "We take privacy seriously. Bento runs exclusively on your computer. It will never send your "
                        "code anywhere.",
                        style={"dim": True},
                    ),
                ],
                processor=Processors.wrap(extra=8),
            )
        ),
        Newline(),
        Echo(
            Text(
                "We’re constantly looking to make Bento better. To that end, we collect limited usage and results "
                "data. To learn more, see bento.dev/privacy.",
                processor=Processors.wrap_link(
                    [Link("bento.dev/privacy", "https://bento.dev/privacy")]
                ),
                style={"dim": True},
            )
        ),
        Newline(),
    )

    invalid_version = Error(
        "~/.bento/config.yml is malformed. Please remove this file and re-run Bento. "
    )

    upgrade = Steps(
        Echo(
            Text(
                "We've made changes to our terms of service. Please review the new terms. If you have any questions "
                "or concerns please reach out via support@r2c.dev.",
                processor=Processors.wrap_link(
                    [Link("support@r2c.dev", "mailto:support@r2c.dev")]
                ),
                style={"dim": True},
            )
        ),
        Newline(),
    )

    prompt = Confirm(
        "Continue and agree to Bento's terms of service and privacy policy?",
        options={"default": True},
    )
    error = Error(
        "Bento did NOT install. Bento beta users must agree to the terms of service to continue. Please reach out to "
        "us at support@r2c.dev with questions or concerns. "
    )


class UpdateGitignore:
    confirm = Confirm(
        content=Multi(
            [
                "Some Bento files should be excluded from version control. Should Bento append them to ",
                Sub(0, style={"bold": True}),
                "?",
            ],
            processor=Processors.wrap(),
        ),
        options={"default": True},
    )
    confirm_yes = Newline()
    confirm_no = Newline()
    update = Progress(
        content=Multi(["Updating ", Sub(0, style={"bold": True, "dim": True})]),
        extra=12,
    )


class SuggestAutocomplete:
    confirm = Confirm(
        Text(
            "Bento supports tab autocompletion. Do you want to add it to your shell profile now?",
            processor=Processors.wrap(),
        ),
        options={"default": True},
    )
    install = Progress(
        content=Multi(
            ["Adding autocompletion to ", Sub(0, style={"bold": True, "dim": True})]
        ),
        extra=12,
    )
    confirm_yes = Newline()
    confirm_no = Echo(
        Multi(
            [
                "\nTo enable autocompletion later, run $ ",
                Text("bento enable autocomplete", style={"bold": True}),
                ".\n",
            ]
        )
    )



from pathlib import Path
from typing import Any, Dict, Set, Tuple

import click

import bento.orchestrator
import bento.result
from bento.context import Context
from bento.decorators import with_metrics
from bento.paths import list_paths
from bento.result import VIOLATIONS_KEY, Baseline
from bento.target_file_manager import TargetFileManager
from bento.util import echo_newline, echo_next_step


def __archive_help(prog_name: str, headless: bool) -> str:
    block_str = (
        "" if headless else " and will not block commits if `autorun` is enabled"
    )
    return f"""
    Suppress current findings.

    By default, only results introduced by currently staged changes will be
    added to the archive (`.bento/archive.json`). Archived findings will
    not appear in future `{prog_name} check` output{block_str}.

    Use `--all` to archive findings in all Git tracked files, not just those
    that are staged:

        $ {prog_name} archive --all [PATHS]

    Optional PATHS can be specified to archive results from specific directories
    or files.

    Archived findings are viewable in `.bento/archive.json`.
    """


@click.command()
@click.option(
    "--all",
    "all_",
    is_flag=True,
    default=False,
    help="Archive findings for all tracked files.",
)
@click.argument("paths", nargs=-1, type=Path, autocompletion=list_paths)
@click.pass_obj
@with_metrics
def archive(context: Context, all_: bool, paths: Tuple[Path, ...]) -> None:
    context.validate_init()

    # Default to no path filter
    path_list = list(paths) if len(paths) > 0 else [context.base_path]

    if not context.is_init:
        if all_:
            click.echo(f"Running archive on all tracked files...\n", err=True)
        else:
            click.echo(f"Running archive on staged files...\n", err=True)

    if context.baseline_file_path.exists():
        with context.baseline_file_path.open() as json_file:
            old_baseline = bento.result.load_baseline(json_file)
            old_hashes = {
                h
                for findings in old_baseline.values()
                for h in findings.get(VIOLATIONS_KEY, {}).keys()
            }
    else:
        old_baseline = {}
        old_hashes = set()

    new_baseline: Dict[str, Dict[str, Dict[str, Any]]] = {}
    tools = context.tools.values()

    target_file_manager = TargetFileManager(
        context.base_path, path_list, not all_, context.ignore_file_path
    )

    baseline: Baseline = {}
    if context.baseline_file_path.exists():
        with context.baseline_file_path.open() as json_file:
            # TODO there's some deconflicting needed with old_baseline, old_hashes, baseline
            baseline = bento.result.json_to_violation_hashes(json_file)

    all_findings, elapsed = bento.orchestrator.orchestrate(
        baseline, target_file_manager, not all_, tools
    )

    n_found = 0
    n_existing = 0
    found_hashes: Set[str] = set()

    for tool_id, vv in all_findings:
        if isinstance(vv, Exception):
            raise vv
        # Remove filtered
        vv = [f for f in vv if not f.filtered]
        n_found += len(vv)
        new_baseline[tool_id] = bento.result.dump_results(vv)
        if tool_id in old_baseline:
            new_baseline[tool_id][VIOLATIONS_KEY].update(
                old_baseline[tool_id][VIOLATIONS_KEY]
            )
        for v in vv:
            h = v.syntactic_identifier_str()
            found_hashes.add(h)
            if h in old_hashes:
                n_existing += 1

    n_new = n_found - n_existing

    context.baseline_file_path.parent.mkdir(exist_ok=True, parents=True)
    with context.baseline_file_path.open("w") as json_file:
        bento.result.write_tool_results(json_file, new_baseline)

    finding_source_text = "in this project" if all_ else "due to staged changes"
    success_str = f"{n_new} finding(s) {finding_source_text} were archived, and will be hidden in future runs."
    if n_existing > 0:
        success_str += f"\nAlso kept {n_existing} existing finding(s)."

    click.echo(success_str, err=True)

    if not context.is_init:
        echo_newline()
        echo_next_step("To view archived results", "cat .bento/archive.json")

    if context.on_exit_exception:
        raise context.on_exit_exception


# TODO: make not monkey with globals
def make_archive(prog_name: str, headless: bool) -> click.Command:
    archive.help = __archive_help(prog_name, headless)
    return archive



import json
import re
from typing import Any, Dict, Iterable, List, Optional, Pattern, Type

import attr
from semantic_version import SimpleSpec

from bento.parser import Parser
from bento.tool import JsonR, output, runner
from bento.util import fetch_line_in_file
from bento.violation import Violation


class ShellcheckParser(Parser[JsonR]):
    def to_violation(self, result: Dict[str, Any]) -> Violation:

        path = self.trim_base(result["file"])
        start_line = result["line"]
        start_col = result["column"]
        message = result.get("message", "")
        check_id = f"SC{result['code']}"

        level = result["level"]
        severity = 0
        if level == "error":
            severity = 2
        elif level == "warning":
            severity = 1

        link = f"https://github.com/koalaman/shellcheck/wiki/{check_id}"
        line_of_code = (
            fetch_line_in_file(self.base_path / path, start_line) or "<no source found>"
        )

        return Violation(
            tool_id=ShellcheckTool.tool_id(),
            check_id=check_id,
            path=path,
            line=start_line,
            column=start_col,
            message=message,
            severity=severity,
            syntactic_context=line_of_code,
            link=link,
        )

    def parse(self, results: JsonR) -> List[Violation]:
        violations: List[Violation] = []
        for check in results:
            violations.append(self.to_violation(check))
        return violations


@attr.s
class ShellcheckTool(runner.Docker, output.Json):
    DOCKER_IMAGE = "koalaman/shellcheck:v0.7.0"
    FILE_NAME_FILTER = re.compile(r".*\.(sh|bash|ksh|dash)$")
    CONTAINER_NAME = "bento-shell-check-daemon"
    TOOL_ID = "shellcheck"
    SHEBANG_PATTERN = re.compile(r"^#!(.*/|.*env +)(sh|bash|ksh)")
    BINARY_SPEC = SimpleSpec("~=0.7.0")
    VERSION_MATCH = re.compile(r"version: (.*)")

    @property
    def binary_executable(self) -> Optional[str]:
        return "shellcheck"

    @property
    def binary_version_pattern(self) -> Pattern:
        return self.VERSION_MATCH

    @property
    def allowed_binary_versions(self) -> SimpleSpec:
        return self.BINARY_SPEC

    @property
    def shebang_pattern(self) -> Optional[Pattern]:
        return self.SHEBANG_PATTERN

    @property
    def binary_command(self) -> List[str]:
        return ["shellcheck"]

    @property
    def parser_type(self) -> Type[Parser]:
        return ShellcheckParser

    @classmethod
    def tool_id(cls) -> str:
        return cls.TOOL_ID

    @classmethod
    def tool_desc(cls) -> str:
        return "Finds bugs in shell scripts (requires Docker)"

    @property
    def file_name_filter(self) -> Pattern:
        return self.FILE_NAME_FILTER

    @property
    def project_name(self) -> str:
        return "Shell"

    @property
    def docker_image(self) -> str:
        return self.DOCKER_IMAGE

    @property
    def remote_code_path(self) -> str:
        return "/mnt/"

    @property
    def docker_command(self) -> List[str]:
        return ["--severity", "info", "-f", "json"]

    def is_allowed_returncode(self, returncode: int) -> bool:
        return returncode == 0 or returncode == 1

    def run(self, files: Iterable[str]) -> JsonR:
        result = self.run_command(files)
        return json.loads(result.stdout)



from typing import Set

import click

from bento.commands.autocomplete import uninstall_autocomplete
from bento.commands.autorun import uninstall_autorun
from bento.commands.ci import uninstall_ci
from bento.config import ToolCommand, get_valid_tools, update_ignores, update_tool_run
from bento.context import Context
from bento.decorators import with_metrics
from bento.util import echo_success


@click.group()
def disable() -> None:
    """
    Turn OFF a Bento feature for this project.

    For example, to disable the `r2c.flask secure-set-cookie` check:

        $ bento disable check r2c.flask secure-set-cookie

    To disable the tool `bandit`:

        $ bento disable tool bandit
    """


@disable.command(
    cls=ToolCommand,
    short_help="Specify a tool to enable.",
    help_summary="Turn OFF a tool.",
)
@click.argument("tool", type=str, nargs=1, autocompletion=get_valid_tools)
@with_metrics
@click.pass_obj
def tool(context: Context, tool: str) -> None:
    """
    Turn OFF a tool.

    Tool-specific configurations are saved, and can be reenabled via `bento enable tool TOOL`.

    Please see `bento disable --help` for more information.
    """
    update_tool_run(context, tool, False)
    echo_success(f"{tool} disabled")


@disable.command(short_help="Specify a check to disable.")
@click.argument("tool", type=str, nargs=1, autocompletion=get_valid_tools)
@click.argument("check", type=str, nargs=1)
@with_metrics
@click.pass_obj
def check(context: Context, tool: str, check: str) -> None:
    """
    Turn OFF a check.

    Visit bento.dev/checks to learn about Bento's specialty checks.
    """

    def add(ignores: Set[str]) -> None:
        ignores.add(check)

    update_ignores(context, tool, add)
    echo_success(f"'{check}' disabled for '{tool}'")


disable.add_command(uninstall_autorun)
disable.add_command(uninstall_autocomplete)
disable.add_command(uninstall_ci)



from abc import ABC, abstractmethod
from typing import Any, Collection, Dict, List, Mapping

import attr

from bento.base_context import BaseContext
from bento.violation import Violation

FindingsMap = Mapping[str, Collection[Violation]]


@attr.s(auto_attribs=True)
class Formatter(ABC):
    """
    Converts tool violations into printable output
    """

    context: BaseContext
    config: Dict[str, Any]

    BOLD = "\033[1m"
    END = "\033[0m"

    @abstractmethod
    def dump(self, findings: FindingsMap) -> Collection[str]:
        """Formats the list of violations for the end user."""
        pass

    @staticmethod
    def path_of(violation: Violation) -> str:
        return violation.path

    @staticmethod
    def by_path(findings: FindingsMap) -> List[Violation]:
        collapsed = (v for violations in findings.values() for v in violations)
        return sorted(collapsed, key=(lambda v: v.path))



#!/usr/bin/env python3
from bento.cli import make_cli
from bento.decorators import echo_exception


@echo_exception
def main() -> None:
    make_cli("bento")(auto_envvar_prefix="BENTO")


@echo_exception
def headless() -> None:
    make_cli("bentoh", headless=True)(
        default_map={"headless": True}, auto_envvar_prefix="BENTO"
    )


if __name__ == "__main__":
    main()



import json
from collections import OrderedDict
from typing import Any, Dict, List, Mapping, Set, TextIO, Union

import attr

from bento.violation import Violation

VIOLATIONS_KEY = "violations"

Hash = str
ToolId = str
Baseline = Dict[str, Set[Hash]]
ToolResults = Mapping[str, Mapping[Hash, Mapping[str, Any]]]


def filtered(
    tool_id: str, output: List[Violation], baseline: Baseline
) -> List[Violation]:
    rejects: Set[Hash] = set(baseline.get(tool_id, {}))
    return [
        attr.evolve(v, filtered=v.syntactic_identifier_str() in rejects) for v in output
    ]


def dump_results(results: List[Violation]) -> Dict[str, Dict[Hash, Dict[str, Any]]]:
    with_hashes: Dict[str, Dict[str, Any]] = OrderedDict(
        sorted(
            ((v.syntactic_identifier_str(), v.to_dict()) for v in results),
            key=(lambda vv: vv[0]),
        )
    )
    return {VIOLATIONS_KEY: with_hashes}


def write_tool_results(stream: TextIO, results: Mapping[ToolId, ToolResults]) -> None:
    json.dump(results, stream, indent=2)


def load_baseline(text: Union[str, TextIO]) -> Mapping[str, ToolResults]:
    parsed = json.loads(text) if isinstance(text, str) else json.load(text)
    return parsed or {}


def json_to_violation_hashes(text: Union[str, TextIO]) -> Baseline:
    parsed = load_baseline(text)
    out = {}
    for (tool_id, r) in parsed.items():
        violations = r[VIOLATIONS_KEY]
        hashes = set(violations.keys()) if violations else set()
        out[tool_id] = hashes
    return out


def to_cache_repr(findings: List[Violation]) -> str:
    as_dict = [attr.asdict(f) for f in findings]
    return json.dumps(as_dict)


def from_cache_repr(text: str) -> List[Violation]:
    parsed = json.loads(text)
    return [Violation(**kwargs) for kwargs in parsed]



from typing import Any, Dict

import bento.util
from bento.base_context import BaseContext
from bento.formatter import base, clippy, histo, json, reporter, stylish

FORMATTERS = {
    "stylish": stylish.Stylish,
    "json": json.Json,
    "clippy": clippy.Clippy,
    "histo": histo.Histo,
    "reporter": reporter.Reporter,
}


def for_name(name: str, context: BaseContext, config: Dict[str, Any]) -> base.Formatter:
    """
    Reflectively instantiates a formatter from a preset name or python identifier

    E.g.
      for_name("stylish", {})
      for_name("bento.formatter.stylish.Stylish", {})
    both return a new instance of the Stylish formatter

    Parameters:
        name (str): The formatter name, as a key of FORMATTERS, or a python fully qualified identifier
        config (dict): The formatter's configuration (formatter-specific)
    """
    tpe = FORMATTERS.get(name.lower(), None)
    if tpe is None:
        tpe = bento.util.for_name(name)
    fmt = tpe(context, config)  # type: ignore
    return fmt


FindingsMap = base.FindingsMap
Formatter = base.Formatter
